"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com


Command line development guidelines
###################################

do's and don'ts
----------------
MUST NOT: sys.exit => use command.CLIException
SHOULD NOT: print( => use logger for messages, only print for final output


Entry points
------------
@command annotation to register new command
"""

import argparse
import asyncio
import contextlib
import dataclasses
import enum
import json
import logging
import os
import shutil
import socket
import sys
import time
import traceback
from argparse import ArgumentParser
from asyncio import ensure_future
from collections import abc
from configparser import ConfigParser
from typing import Optional

import click
from tornado.httpclient import AsyncHTTPClient
from tornado.ioloop import IOLoop

import inmanta.compiler as compiler
from inmanta import const, module, moduletool, protocol, tracing, util
from inmanta.agent import config as agent_config
from inmanta.ast import CompilerException, Namespace
from inmanta.ast import type as inmanta_type
from inmanta.command import CLIException, Commander, ShowUsageException, command
from inmanta.compiler import do_compile
from inmanta.config import Config, Option
from inmanta.const import ALL_LOG_CONTEXT_VARS, EXIT_START_FAILED, LOG_CONTEXT_VAR_ENVIRONMENT
from inmanta.export import cfg_env
from inmanta.logging import InmantaLoggerConfig, _is_on_tty
from inmanta.server import config as opt
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server.services.databaseservice import initialize_database_connection_pool
from inmanta.server.services.metricservice import MetricsService
from inmanta.signals import safe_shutdown, setup_signal_handlers
from inmanta.util import get_compiler_version
from inmanta.warnings import WarningsManager

LOGGER = logging.getLogger("inmanta")


def server_parser_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[argparse.ArgumentParser]) -> None:
    parser.add_argument(
        "--db-wait-time",
        type=int,
        dest="db_wait_time",
        help="Maximum time in seconds the server will wait for the database to be up before starting. "
        "A value of 0 means the server will not wait. If set to a negative value, the server will wait indefinitely.",
    )
    parser.add_argument(
        "--compatibility-file",
        type=str,
        dest="compatibility_file",
        help="Path to the compatibility.json file. The constraints defined in the `python_package_constraints` field will be "
        "enforced both during project install and during agent install. For more information about this file, please refer to "
        "the compatibility page in the Inmanta documentation.",
    )


@command("server", help_msg="Start the inmanta server", parser_config=server_parser_config, component="server")
def start_server(options: argparse.Namespace) -> None:
    if options.config_file and not os.path.exists(options.config_file):
        LOGGER.warning("Config file %s doesn't exist", options.config_file)

    if options.config_dir and not os.path.isdir(options.config_dir):
        LOGGER.warning("Config directory %s doesn't exist", options.config_dir)

    if options.db_wait_time is not None:
        Config.set("database", "wait_time", str(options.db_wait_time))

    if options.compatibility_file is not None:
        Config.set("server", "compatibility_file", str(options.compatibility_file))

    tracing.configure_logfire("server")
    util.ensure_event_loop()

    ibl = InmantaBootloader()

    setup_signal_handlers(ibl.stop)

    ioloop = IOLoop.current()

    # handle startup exceptions
    def _handle_startup_done(fut: asyncio.Future) -> None:
        if fut.cancelled():
            safe_shutdown(ioloop, ibl.stop)
        else:
            exc = fut.exception()
            if exc is not None:
                LOGGER.exception("Server setup failed", exc_info=exc)
                traceback.print_exception(type(exc), exc, exc.__traceback__)
                safe_shutdown(ioloop, ibl.stop)
            else:
                LOGGER.info("Server startup complete")

    ensure_future(ibl.start()).add_done_callback(_handle_startup_done)

    ioloop.start()
    LOGGER.info("Server shutdown complete")
    if not ibl.started:
        exit(EXIT_START_FAILED)


@command("scheduler", help_msg="Start the resource scheduler", component="scheduler")
def start_scheduler(options: argparse.Namespace) -> None:
    """
    Start the new agent with the Resource Scheduler
    """
    from inmanta.agent import agent_new

    # The call to configure() should be done as soon as possible.
    # If an AsyncHTTPClient is started before this call, the max_client
    # will not be taken into account.
    max_clients: Optional[int] = Config.get(section="agent_rest_transport", name="max_clients")

    if max_clients:
        AsyncHTTPClient.configure(None, max_clients=max_clients)

    tracing.configure_logfire("scheduler")

    util.ensure_event_loop()
    a = agent_new.Agent()

    async def start() -> None:
        await initialize_database_connection_pool(
            database_host=opt.db_host.get(),
            database_port=opt.db_port.get(),
            database_name=opt.db_name.get(),
            database_username=opt.db_username.get(),
            database_password=opt.db_password.get(),
            create_db_schema=True,
            connection_pool_min_size=agent_config.scheduler_db_connection_pool_min_size.get(),
            connection_pool_max_size=agent_config.scheduler_db_connection_pool_max_size.get(),
            connection_timeout=agent_config.scheduler_db_connection_timeout.get(),
        )
        # also report metrics if this is relevant
        metrics_reporter = MetricsService(
            extra_tags={"component": "scheduler", "environment": str(agent_config.environment.get())}
        )
        metrics_reporter.start_metric_reporters()
        await a.start()

    LOGGER.info("Agent with Resource scheduler starting now")
    setup_signal_handlers(a.stop)
    IOLoop.current().add_callback(start)
    IOLoop.current().start()
    LOGGER.info("Agent with Resource scheduler Shutdown complete")


class ExperimentalFeatureFlags:
    """
    Class to expose feature flag configs as options in a uniform matter
    """

    def __init__(self) -> None:
        self.metavar_to_option: dict[str, Option[bool]] = {}

    def _get_name(self, option: Option[bool]) -> str:
        return f"flag_{option.name}"

    def add(self, option: Option[bool]) -> None:
        """Add an option to the set of feature flags"""
        self.metavar_to_option[self._get_name(option)] = option

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        """Add all feature flag options to the argument parser"""
        for metavar, option in self.metavar_to_option.items():
            parser.add_argument(
                f"--experimental-{option.name}",
                dest=metavar,
                help=option.documentation,
                action="store_true",
                default=False,
            )

    def read_options_to_config(self, options: argparse.Namespace) -> None:
        """
        This method takes input from the commandline parser
        and sets the appropriate feature flag config based
        on the parsed command line arguments

        :param options: the options, as parsed by argparse.
        """
        for metavar, option in self.metavar_to_option.items():
            value = getattr(options, metavar, False)
            if value:
                option.set("true")


def compiler_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    """
    Configure the compiler of the export function
    """
    parser.add_argument("-e", dest="environment", help="The environment to compile this model for")
    parser.add_argument(
        "-X",
        "--extended-errors",
        dest="errors",
        help="Show stack traces for compile errors",
        action="store_true",
        default=argparse.SUPPRESS,
    )
    parser.add_argument("--server_address", dest="server", help="The address of the server hosting the environment")
    parser.add_argument("--server_port", dest="port", help="The port of the server hosting the environment")
    parser.add_argument("--username", dest="user", help="The username of the server")
    parser.add_argument("--password", dest="password", help="The password of the server")
    parser.add_argument("--ssl", help="Enable SSL", action="store_true", default=False)
    parser.add_argument("--ssl-ca-cert", dest="ca_cert", help="Certificate authority for SSL")
    parser.add_argument(
        "--export-compile-data",
        dest="export_compile_data",
        help="Export structured json containing compile data such as occurred errors.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--export-compile-data-file",
        dest="export_compile_data_file",
        help="File to export compile data to. If omitted %s is used." % compiler.config.default_compile_data_file,
    )
    parser.add_argument(
        "--no-cache",
        dest="feature_compiler_cache",
        help="Disable caching of compiled CF files",
        action="store_false",
        default=True,
    )
    parser.add_argument(
        "--experimental-data-trace",
        dest="datatrace",
        help="Experimental data trace tool useful for debugging",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--experimental-dataflow-graphic",
        dest="dataflow_graphic",
        help="Experimental graphic data flow visualization",
        action="store_true",
        default=False,
    )

    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")
    moduletool.add_deps_check_arguments(parser)


@command(
    "compile",
    help_msg="Compile the project to a configuration model",
    parser_config=compiler_config,
    require_project=True,
    component="compiler",
)
def compile_project(options: argparse.Namespace) -> None:
    if options.environment is not None:
        Config.set("config", "environment", options.environment)

    if options.server is not None:
        Config.set("compiler_rest_transport", "host", options.server)

    if options.port is not None:
        Config.set("compiler_rest_transport", "port", options.port)

    if options.user is not None:
        Config.set("compiler_rest_transport", "username", options.user)

    if options.password is not None:
        Config.set("compiler_rest_transport", "password", options.password)

    if options.ssl:
        Config.set("compiler_rest_transport", "ssl", "true")

    if options.ca_cert is not None:
        Config.set("compiler_rest_transport", "ssl-ca-cert-file", options.ca_cert)

    if options.export_compile_data is True:
        Config.set("compiler", "export_compile_data", "true")

    if options.export_compile_data_file is not None:
        Config.set("compiler", "export_compile_data_file", options.export_compile_data_file)

    if options.feature_compiler_cache is False:
        Config.set("compiler", "cache", "false")

    if options.datatrace is True:
        Config.set("compiler", "datatrace_enable", "true")

    if options.dataflow_graphic is True:
        Config.set("compiler", "dataflow_graphic_enable", "true")

    module.Project.get(options.main_file)

    with tracing.span("compile"):
        summary_reporter = CompileSummaryReporter()
        if options.profile:
            import cProfile
            import pstats

            with summary_reporter.compiler_exception.capture():
                cProfile.runctx("do_compile()", globals(), {}, "run.profile")
            p = pstats.Stats("run.profile")
            p.strip_dirs().sort_stats("time").print_stats(20)
        else:
            t1 = time.time()
            with summary_reporter.compiler_exception.capture():
                do_compile()
            LOGGER.debug("The entire compile command took %0.03f seconds", time.time() - t1)

        summary_reporter.print_summary_and_exit(show_stack_traces=options.errors)


@command("list-commands", help_msg="Print out an overview of all commands", add_verbose_flag=False)
def list_commands(options: argparse.Namespace) -> None:
    print("The following commands are available:")
    for cmd, info in Commander.commands().items():
        print(" {}: {}".format(cmd, info["help"]))


def help_parser_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    parser.add_argument("subcommand", help="Output help for a particular subcommand", nargs="?", default=None)


@command("help", help_msg="show a help message and exit", parser_config=help_parser_config, add_verbose_flag=False)
def help_command(options: argparse.Namespace) -> None:
    if options.subcommand is None:
        cmd_parser().print_help()
    else:
        subc = options.subcommand
        parser = cmd_parser()
        parser.parse_args([subc, "-h"])
    sys.exit(0)


@command(
    "modules",
    help_msg="Subcommand to manage modules",
    parser_config=moduletool.ModuleTool.modules_parser_config,
    aliases=["module"],
)
def modules(options: argparse.Namespace) -> None:
    tool = moduletool.ModuleTool()
    tool.execute(options.cmd, options)


@command("project", help_msg="Subcommand to manage the project", parser_config=moduletool.ProjectTool.parser_config)
def project(options: argparse.Namespace) -> None:
    tool = moduletool.ProjectTool()
    tool.execute(options.cmd, options)


def export_parser_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    """
    Configure the compiler of the export function
    """
    parser.add_argument("-g", dest="depgraph", help="Dump the dependency graph", action="store_true")
    parser.add_argument(
        "-j",
        dest="json",
        help="Do not submit to the server but only store the json that would have been " "submitted in the supplied file",
    )
    parser.add_argument("-e", dest="environment", help="The environment to compile this model for")
    parser.add_argument("-d", dest="deploy", help="Trigger a deploy for the exported version", action="store_true")
    parser.add_argument(
        "--full",
        dest="full_deploy",
        help="Make the agents execute a full deploy instead of an incremental deploy. "
        "Should be used together with the -d option",
        action="store_true",
        default=False,
    )
    parser.add_argument("-m", dest="model", help="Also export the complete model", action="store_true", default=False)
    parser.add_argument("--server_address", dest="server", help="The address of the server to submit the model to")
    parser.add_argument("--server_port", dest="port", help="The port of the server to submit the model to")
    parser.add_argument("--token", dest="token", help="The token to auth to the server")
    parser.add_argument("--ssl", help="Enable SSL", action=argparse.BooleanOptionalAction, default=None)
    parser.add_argument("--ssl-ca-cert", dest="ca_cert", help="Certificate authority for SSL")
    parser.add_argument(
        "-X",
        "--extended-errors",
        dest="errors",
        help="Show stack traces for compile errors",
        action="store_true",
        default=argparse.SUPPRESS,
    )
    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")
    parser.add_argument(
        "--metadata",
        dest="metadata",
        help="JSON metadata why this compile happened. If a non-json string is "
        "passed it is used as the 'message' attribute in the metadata.",
        default=None,
    )
    parser.add_argument(
        "--model-export",
        dest="model_export",
        help="Export the configuration model to the server as metadata.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--export-plugin",
        dest="export_plugin",
        help="Only use this export plugin. This option also disables the execution of the plugins listed in "
        "the configuration file in the export setting.",
        default=None,
    )

    parser.add_argument(
        "--export-compile-data",
        dest="export_compile_data",
        help="Export structured json containing compile data such as occurred errors.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--export-compile-data-file",
        dest="export_compile_data_file",
        help="File to export compile data to. If omitted %s is used." % compiler.config.default_compile_data_file,
    )
    parser.add_argument(
        "--no-cache",
        dest="feature_compiler_cache",
        help="Disable caching of compiled CF files",
        action="store_false",
        default=True,
    )
    parser.add_argument(
        "--partial",
        dest="partial_compile",
        help=(
            "Execute a partial export. Does not upload new Python code to the server: it is assumed to be unchanged since the"
            " last full export. Multiple partial exports for disjunct resource sets may be performed concurrently but not"
            " concurrent with a full export. When used in combination with the ``--json`` option, 0 is used as a placeholder "
            "for the model version."
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--delete-resource-set",
        dest="delete_resource_set",
        help="Remove a resource set as part of a partial compile. This option can be provided multiple times and should always "
        "be used together with the --partial option. Sets can also be marked for deletion via the INMANTA_REMOVED_SET_ID "
        "env variable as a space separated list of set ids to remove.",
        action="append",
    )
    parser.add_argument(
        "--soft-delete",
        dest="soft_delete",
        help="This flag prevents the deletion of resource sets (marked for deletion via the ``--delete-resource-set`` cli "
        "option or the INMANTA_REMOVED_SET_ID env variable) that contain resources that are currently being exported.",
        action="store_true",
        default=False,
    )
    moduletool.add_deps_check_arguments(parser)


@command(
    "export",
    help_msg="Export the configuration",
    parser_config=export_parser_config,
    require_project=True,
    component="compiler",
)
def export(options: argparse.Namespace) -> None:
    resource_sets_to_remove: set[str] = set(options.delete_resource_set) if options.delete_resource_set else set()

    if const.INMANTA_REMOVED_SET_ID in os.environ:
        removed_sets = set(os.environ[const.INMANTA_REMOVED_SET_ID].split())

        resource_sets_to_remove.update(removed_sets)

    if not options.partial_compile and resource_sets_to_remove:
        raise CLIException(
            "A full export was requested but resource sets were marked for deletion (via the --delete-resource-set cli "
            "option or the INMANTA_REMOVED_SET_ID env variable). Deleting a resource set can only be performed during a "
            "partial export. To trigger a partial export, use the --partial option.",
            exitcode=1,
        )

    if options.environment is not None:
        Config.set("config", "environment", options.environment)

    if options.server is not None:
        Config.set("compiler_rest_transport", "host", options.server)

    if options.port is not None:
        Config.set("compiler_rest_transport", "port", options.port)

    if options.token is not None:
        Config.set("compiler_rest_transport", "token", options.token)

    if options.ssl is not None:
        Config.set("compiler_rest_transport", "ssl", f"{options.ssl}".lower())

    if options.ca_cert is not None:
        Config.set("compiler_rest_transport", "ssl-ca-cert-file", options.ca_cert)

    if options.export_compile_data is True:
        Config.set("compiler", "export_compile_data", "true")

    if options.export_compile_data_file is not None:
        Config.set("compiler", "export_compile_data_file", options.export_compile_data_file)

    if options.feature_compiler_cache is False:
        Config.set("compiler", "cache", "false")

    tracing.configure_logfire("compiler")

    # try to parse the metadata as json. If a normal string, create json for it.
    if options.metadata is not None and len(options.metadata) > 0:
        try:
            metadata = json.loads(options.metadata)
        except json.decoder.JSONDecodeError:
            metadata = {"message": options.metadata}
    else:
        metadata = {"message": "Manual compile on the CLI by user"}

    if "cli-user" not in metadata and "USERNAME" in os.environ:
        metadata["cli-user"] = os.environ["USERNAME"]

    if "hostname" not in metadata:
        metadata["hostname"] = socket.gethostname()

    if "type" not in metadata:
        metadata["type"] = "manual"

    module.Project.get(options.main_file)

    from inmanta.export import Exporter  # noqa: H307

    with tracing.span("compiler"):
        summary_reporter = CompileSummaryReporter()

        types: Optional[dict[str, inmanta_type.Type]]
        scopes: Optional[Namespace]

        t1 = time.time()
        with summary_reporter.compiler_exception.capture():
            try:
                (types, scopes) = do_compile()
            except Exception:
                types, scopes = (None, None)
                raise

    # Even if the compile failed we might have collected additional data such as unknowns. So
    # continue the export
    with tracing.span("exporter"):
        export = Exporter(options)
        with summary_reporter.exporter_exception.capture():
            results = export.run(
                types,
                scopes,
                metadata=metadata,
                model_export=options.model_export,
                export_plugin=options.export_plugin,
                partial_compile=options.partial_compile,
                resource_sets_to_remove=list(resource_sets_to_remove),
            )

        if not summary_reporter.is_failure() and options.deploy:
            version = results[0]
            conn = protocol.SyncClient("compiler")
            LOGGER.info("Triggering deploy for version %d" % version)
            tid = cfg_env.get()
            agent_trigger_method = const.AgentTriggerMethod.get_agent_trigger_method(options.full_deploy)
            conn.release_version(tid, version, True, agent_trigger_method)

    LOGGER.debug("The entire export command took %0.03f seconds", time.time() - t1)
    summary_reporter.print_summary_and_exit(show_stack_traces=options.errors)


def validate_logging_config_parser_config(
    parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]
) -> None:
    """
    Config parser for the validate-logging-config command.
    """
    parser.add_argument(
        "-e",
        dest="environment",
        help="The environment id to be used as context variable in logging config templates. If not specified,"
        " 0c111d30-feaf-4f5b-b2d6-83d589480a4a will be used.",
        default="0c111d30-feaf-4f5b-b2d6-83d589480a4a",
    )

    sub_parsers = parser.add_subparsers(title="subcommand", dest="cmd")
    for component_name in ["server", "scheduler", "compiler"]:
        sub_parser = sub_parsers.add_parser(
            component_name, help=f"Validate the logging config for the {component_name}", parents=parent_parsers
        )
        sub_parser.set_defaults(component=component_name)


@command(
    "validate-logging-config",
    help_msg="This command loads the logging config like other CLI commands would (taking into account"
    " the precedence rules for logging configuration) and produces log lines. It serves as a tool to validate whether"
    " a logging config file is syntactically correct and behaves as expected. Optionally, a sub-command can be specified"
    " to indicate the component for which the logging config file should be loaded.",
    parser_config=validate_logging_config_parser_config,
)
def validate_logging_config(options: argparse.Namespace) -> None:
    logging_config = InmantaLoggerConfig.get_current_instance()
    if logging_config.logging_config_source is None:
        raise Exception("No logging configuration found.")
    print(f"Using logging config from {logging_config.logging_config_source.source()}", file=sys.stderr)
    env_id = options.environment
    logger_and_message = [
        (logging.getLogger("inmanta.protocol.rest.server"), "Log line from Inmanta server"),
        (logging.getLogger("inmanta.server.services.compilerservice"), "Log line from compiler service"),
        (logging.getLogger(const.NAME_RESOURCE_ACTION_LOGGER).getChild(env_id), "Log line for resource action log"),
        (logging.getLogger("inmanta_lsm.callback"), "Log line from callback"),
        (logging.getLogger("inmanta.scheduler"), "Log line from the resource scheduler"),
        (logging.getLogger(const.LOGGER_NAME_EXECUTOR), "Log line from the executor"),
        (logging.getLogger(f"{const.LOGGER_NAME_EXECUTOR}.test"), "Log line from a sub-logger of the executor"),
        (logging.getLogger("performance"), "Performance log line"),
        (logging.getLogger("inmanta.warnings"), "Warning log line"),
        (logging.getLogger("tornado.access"), "tornado access log"),
        (logging.getLogger("tornado.general"), "tornado general log"),
    ]
    log_levels = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]
    print(
        "Each of the log lines mentioned below will be emitted at the following log levels:"
        f" {[logging.getLevelName(level) for level in log_levels]}:",
        file=sys.stderr,
    )
    for logger, msg in logger_and_message:
        print(f" * Emitting log line '{msg} at level <LEVEL>' using logger '{logger.name}'", file=sys.stderr)
        for log_level in log_levels:
            logger.log(log_level, f"{msg} at level {logging.getLevelName(log_level)}")


class Color(enum.Enum):
    RED = "red"
    GREEN = "green"


@dataclasses.dataclass
class ExceptionCollector:
    """
    This class defines a context manager that captures any unhandled exception raised within the context.
    """

    exception: Optional[Exception] = None

    def has_exception(self) -> bool:
        return self.exception is not None

    @contextlib.contextmanager
    def capture(self) -> abc.Iterator["ExceptionCollector"]:
        """
        Record any exceptions raised within the context manager. The exception will not be re-raised.
        """
        try:
            yield self
        except Exception as e:
            self.exception = e


class CompileSummaryReporter:
    """
    Contains the logic to print a summary at the end of the `inmanta compile` or `inmanta export`
    command that provides an overview on whether the command was successful or not.
    """

    def __init__(self) -> None:
        self.compiler_exception: ExceptionCollector = ExceptionCollector()
        self.exporter_exception: ExceptionCollector = ExceptionCollector()

    def is_failure(self) -> bool:
        """
        Return true iff an exception has occurred during the compile or export stage.
        """
        return self.compiler_exception.has_exception() or self.exporter_exception.has_exception()

    def _get_global_status(self) -> str:
        """
        Return the global status of the run.
        """
        if self.compiler_exception.has_exception():
            return "COMPILATION FAILURE"
        elif self.exporter_exception.has_exception():
            return "EXPORT FAILURE"
        else:
            return "SUCCESS"

    def _get_header(self, header_text: str) -> str:
        """
        Return a header for the summary with the given header_text.
        """
        terminal_width = shutil.get_terminal_size()[0]
        minimal_header = f"= {header_text.upper()} ="
        length_minimal_header = len(minimal_header)
        if terminal_width <= length_minimal_header:
            return minimal_header
        else:
            nr_equals_signs_to_add_each_side = int((terminal_width - length_minimal_header) / 2)
            extra_equals_signs_each_side = "=" * nr_equals_signs_to_add_each_side
            return f"{extra_equals_signs_each_side}{minimal_header}{extra_equals_signs_each_side}"

    def _get_exception_to_report(self) -> Exception:
        """
        Return the exception that should be reported in the summary. Compiler exceptions take precedence
        over exporter exceptions because they happen first.
        """
        assert self.is_failure()
        if self.compiler_exception.has_exception():
            assert self.compiler_exception.exception is not None
            return self.compiler_exception.exception
        else:
            assert self.exporter_exception.exception is not None
            return self.exporter_exception.exception

    def _get_error_message(self) -> str:
        """
        Return the error message associated with `self._get_exception_to_report()`.
        """
        exc = self._get_exception_to_report()
        if isinstance(exc, CompilerException):
            error_message = exc.format_trace(indent="  ").strip("\n")
            # Add explainer text if any
            from inmanta.compiler.help.explainer import ExplainerFactory

            helpmsg = ExplainerFactory().explain_and_format(exc, plain=not _is_on_tty())
            if helpmsg is not None:
                helpmsg = helpmsg.strip("\n")
                return f"{error_message}\n\n{helpmsg}"
        else:
            error_message = str(exc).strip("\n")
        return f"Error: {error_message}"

    def _get_stack_trace(self) -> str:
        """
        Return the stack trace associated with `self._get_exception_to_report()`.
        """
        exc = self._get_exception_to_report()
        return "".join(traceback.format_exception(None, value=exc, tb=exc.__traceback__)).strip("\n")

    def _print_to_stderr(self, text: str = "", bold: bool = False, **kwargs: object) -> None:
        """
        Prints the given text to stderr with the given styling requirements. On a tty the text
        is printed in green in case of success and in red in case of a failure.
        """
        if _is_on_tty():
            color = Color.RED if self.is_failure() else Color.GREEN
            text = click.style(text, fg=color.value, bold=bold)
        print(text, file=sys.stderr, **kwargs)

    def print_summary(self, show_stack_traces: bool) -> None:
        """
        Print the summary of the compile run.
        """
        self._print_to_stderr()
        if show_stack_traces and self.is_failure():
            self._print_to_stderr(text=self._get_header(header_text="EXCEPTION TRACE"))
            self._print_to_stderr(text=self._get_stack_trace(), end="\n\n")

        self._print_to_stderr(text=self._get_header(header_text=self._get_global_status()))
        if self.is_failure():
            self._print_to_stderr(text=self._get_error_message())

    def print_summary_and_exit(self, show_stack_traces: bool) -> None:
        """
        Print the compile summary and exit with a 0 status code in case of success or 1 in case of failure.
        """
        self.print_summary(show_stack_traces)
        exit(1 if self.is_failure() else 0)


def cmd_parser() -> argparse.ArgumentParser:
    # create the argument compiler

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", action="store_true", dest="profile", help="Profile this run of the program")
    parser.add_argument("-c", "--config", dest="config_file", help="Use this config file", default=None)
    parser.add_argument(
        "--config-dir",
        dest="config_dir",
        help="The directory containing the Inmanta configuration files",
        default="/etc/inmanta/inmanta.d",
    )
    parser.add_argument("--log-file", dest="log_file", help="Path to the logfile")
    parser.add_argument(
        "--logging-config",
        dest="logging_config",
        help="The path to the configuration file for the logging framework. This is a YAML file that follows "
        "the dictionary-schema accepted by logging.config.dictConfig(). All other log-related configuration "
        "arguments will be ignored when this argument is provided.",
    )
    parser.add_argument(
        "--log-file-level",
        dest="log_file_level",
        choices=["0", "1", "2", "3", "4", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE"],
        default="INFO",
        help="Log level for messages going to the logfile: 0=ERROR, 1=WARNING, 2=INFO, 3=DEBUG",
    )
    parser.add_argument("--timed-logs", dest="timed", help="Add timestamps to logs", action="store_true")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Log level for messages going to the console. Default is warnings only. "
        "When used in combination with a logging config file, it will force a cli logger to be added to the config."
        "-v warning, -vv info, -vvv debug and -vvvv trace",
    )
    parser.add_argument(
        "--warnings",
        dest="warnings",
        choices=["warn", "ignore", "error"],
        default="warn",
        help="The warning behaviour. Must be one of 'warn', 'ignore', 'error'",
    )
    parser.add_argument(
        "-X", "--extended-errors", dest="errors", help="Show stack traces for errors", action="store_true", default=False
    )
    parser.add_argument(
        "--version",
        action="store_true",
        dest="inmanta_version",
        help="Show the version of the installed Inmanta product and the version of its subcomponents",
        default=False,
        required=False,
    )
    parser.add_argument(
        "--keep-logger-names",
        dest="keep_logger_names",
        help="Display the log messages using the name of the logger that created the log messages when running the compiler.",
        action="store_true",
        default=False,
    )

    verbosity_parser = argparse.ArgumentParser(add_help=False)
    verbosity_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=argparse.SUPPRESS,
        help="Log level for messages going to the console. Default is warnings,"
        "-v warning, -vv info, -vvv debug and -vvvv trace",
    )

    subparsers = parser.add_subparsers(title="commands")
    for cmd_name, cmd_options in Commander.commands().items():
        parent_parsers: list[argparse.ArgumentParser] = []
        if cmd_options["add_verbose_flag"]:
            parent_parsers.append(verbosity_parser)
        cmd_subparser = subparsers.add_parser(
            cmd_name, help=cmd_options["help"], aliases=cmd_options["aliases"], parents=parent_parsers
        )
        if cmd_options["parser_config"] is not None:
            cmd_options["parser_config"](cmd_subparser, parent_parsers)
        cmd_subparser.set_defaults(func=cmd_options["function"])
        cmd_subparser.set_defaults(component=cmd_options["component"])
        cmd_subparser.set_defaults(require_project=cmd_options["require_project"])

    return parser


def default_log_config_parser(parser: ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    parser.add_argument(
        "--component",
        dest="config_for_component",
        choices=["server", "scheduler", "compiler"],
        help="The component for which the logging configuration has to be generated.",
    )
    parser.add_argument(
        "output_file",
        help="The file where the logging config should be saved. For the scheduler component, this file must end with a .tmpl"
        " suffix, because a logging configuration template will be generated.",
    )


@command(
    "output-default-logging-config",
    help_msg="Write the default log config for the provided component to file",
    parser_config=default_log_config_parser,
)
def default_logging_config(options: argparse.Namespace) -> None:
    if os.path.exists(options.output_file):
        raise Exception(f"The requested output location already exists: {options.output_file}")
    if options.config_for_component == "scheduler" and not options.output_file.endswith(".tmpl"):
        raise Exception(
            "The config being generated will be a template, but the given filename doesn't end with the .tmpl suffix."
        )

    # Because we want to have contex vars in the files,
    #   but the file can also contain other f-string formatters, this is a bit tricky.
    # We want to be able to
    #   1. replace all env_vars with a template `{env_var}`
    #   2. if we form a template, escape all existing '{' into '{{' and `}` into `}}`
    # What we do is:
    #   1. set all env vars to {placeholder}{var}{placeholder}
    #   2. if we detect the placeholder
    #   3. do escaping
    #   4. replace placeholder with {placeholder}{var}{placeholder} with {{{var}}}

    # Al possible context vars:
    context_vars = ALL_LOG_CONTEXT_VARS

    # 1. Replace variables by placeholder
    # Should be safe as we only expect default configs
    # That don't contain this placeholder
    place_holder = "__PLACE_HOLDER__"

    context = {var: f"{place_holder}{var}{place_holder}" for var in context_vars}

    # Force TTY so that this command outputs the same config when piping to a file
    original_force_tty = os.environ.get(const.ENVIRON_FORCE_TTY, None)
    if original_force_tty is None:
        os.environ[const.ENVIRON_FORCE_TTY] = "yes"
    try:
        component_config = InmantaLoggerConfig(stream=sys.stdout, no_install=True)
        component_config.apply_options(options, options.config_for_component, context)

        if options.config_for_component == "server":
            # Upgrade with extensions
            ibl = InmantaBootloader()
            ibl.start_loggers_for_extensions(component_config)

        assert component_config._loaded_config is not None  # make mypy happy
        raw_dump = component_config._loaded_config.to_string()

        # 2. if we detect the placeholder
        if place_holder in raw_dump:
            # 3. escape all '{' and '}'
            # i.e. we could be a template of a template
            raw_dump = raw_dump.replace("{", "{{")
            raw_dump = raw_dump.replace("}", "}}")
            # 4. replace placeholder with `{`
            for context_var in context_vars:
                raw_dump = raw_dump.replace(f"{place_holder}{context_var}{place_holder}", "{" + context_var + "}")

        with open(options.output_file, "w") as fh:
            fh.write(raw_dump)
    finally:
        # Revert this env var back to its original state
        if original_force_tty is None:
            del os.environ[const.ENVIRON_FORCE_TTY]


def print_versions_installed_components_and_exit() -> None:
    # coroutine to make sure event loop is running for server slices
    async def print_status() -> None:
        bootloader = InmantaBootloader()
        app_context = bootloader.load_slices()
        product_metadata = app_context.get_product_metadata()
        extension_statuses = app_context.get_extension_statuses()

        if product_metadata.version:
            print(f"{product_metadata.product} ({product_metadata.edition}): {product_metadata.version}")
        else:
            print(f"{product_metadata.product} ({product_metadata.edition}): version unknown")
        print(f"Compiler version: {get_compiler_version()}")
        if extension_statuses:
            print("Extensions:")
            for ext_status in extension_statuses:
                print(f"    * {ext_status.name}: {ext_status.version}")
        else:
            print("Extensions: No extensions found")

        await bootloader.stop()

    asyncio.run(print_status())
    sys.exit(0)


def app() -> None:
    """
    Run the compiler
    """
    # Bootstrap log config
    log_config = InmantaLoggerConfig.get_instance()
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)

    # do an initial load of known config files to build the libdir path
    Config.load_config()
    parser = cmd_parser()
    options, other = parser.parse_known_args()
    options.other = other

    if options.config_file and not os.path.exists(options.config_file):
        LOGGER.warning("Config file %s doesn't exist", options.config_file)

    # Load the configuration
    Config.load_config(min_c_config_file=options.config_file, config_dir=options.config_dir)

    # Collect potential log context
    log_context: dict[str, str] = {}
    env: str | None = None
    if hasattr(options, "environment"):
        env = options.environment
    if not env:
        env = str(agent_config.environment.get())
    if env:
        log_context[LOG_CONTEXT_VAR_ENVIRONMENT] = env

    # Log config
    component = options.component if hasattr(options, "component") else None
    log_config.apply_options(options, component=component, context=log_context)
    logging.captureWarnings(True)

    if options.inmanta_version:
        print_versions_installed_components_and_exit()

    if options.warnings is not None:
        Config.set("warnings", "default", options.warnings)

    config = Config.get()
    assert isinstance(config, ConfigParser)
    WarningsManager.apply_config(config["warnings"] if "warnings" in config else None)

    # start the command
    if not hasattr(options, "func"):
        # show help
        parser.print_usage()
        return

    def report(e: BaseException) -> None:
        if not options.errors:
            if isinstance(e, CompilerException):
                print(e.format_trace(indent="  "), file=sys.stderr)
            else:
                print(str(e), file=sys.stderr)
        else:
            sys.excepthook(*sys.exc_info())

        if isinstance(e, CompilerException):
            from inmanta.compiler.help.explainer import ExplainerFactory

            helpmsg = ExplainerFactory().explain_and_format(e, plain=not _is_on_tty())
            if helpmsg is not None:
                print(helpmsg)

    # if a traceparent is provided, restore the context
    with tracing.attach_context({const.TRACEPARENT: os.environ[const.TRACEPARENT]} if const.TRACEPARENT in os.environ else {}):
        try:
            options.func(options)
        except ShowUsageException as e:
            print(e.args[0], file=sys.stderr)
            parser.print_usage()
        except CLIException as e:
            report(e)
            sys.exit(e.exitcode)
        except Exception as e:
            report(e)
            sys.exit(1)
        except KeyboardInterrupt as e:
            report(e)
            sys.exit(1)
        sys.exit(0)


if __name__ == "__main__":
    app()



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging
import re
import typing
import uuid
from collections.abc import Iterable, Iterator, Sequence
from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

import inmanta.ast
import inmanta.util
from inmanta import const, references
from inmanta.execute import proxy, util
from inmanta.stable_api import stable_api
from inmanta.types import JsonType, ResourceIdStr, ResourceVersionIdStr

if TYPE_CHECKING:
    from inmanta import export
    from inmanta.agent import handler
    from inmanta.data import ResourceAction
    from inmanta.execute import runtime

LOGGER = logging.getLogger(__name__)


class ResourceException(Exception):
    pass


@stable_api
class resource:  # noqa: N801
    """
    A decorator that registers a new resource. The decorator must be applied to classes that inherit from
    :class:`~inmanta.resources.Resource`

    :param name: The name of the entity in the configuration model it creates a resource from. For example
        :inmanta:entity:`std::testing::NullResource`
    :param id_attribute: The attribute of `this` resource that uniquely identifies a resource on a logical agent.
        This attribute can be mapped.
    :param agent: This string indicates how the agent of this resource is determined. This string points to an attribute,
        but it can navigate relations (this value cannot be mapped). For example, the agent argument could be ``host.name``.
    """

    # The _resources dict is accessed by the compile function in pytest-inmanta.
    # see https://github.com/inmanta/pytest-inmanta/pull/381
    _resources: dict[str, tuple[type["Resource"], dict[str, str]]] = {}

    def __init__(self, name: str, id_attribute: str, agent: str):
        if not isinstance(agent, str):
            raise ResourceException(f"The agent parameter has to be a string, got {agent} of type {type(agent)}")
        self._cls_name = name
        self._options = {"agent": agent, "name": id_attribute}

    def __call__[R: Resource](self, cls: type[R]) -> type[R]:
        """
        The wrapping
        """
        if self._cls_name in resource._resources:
            LOGGER.info("Reloading module %s" % self._cls_name)
            del resource._resources[self._cls_name]

        resource._resources[self._cls_name] = (cls, self._options)
        return cls

    @classmethod
    def add_resource(cls, cls_name: str, resource_type: type["Resource"], options: dict[str, str]) -> None:
        cls._resources[cls_name] = (resource_type, options)

    @classmethod
    def validate(cls) -> None:
        fq_name_resource_decorator = f"{cls.__module__}.{cls.__name__}"
        fq_name_resource_class = f"{Resource.__module__}.{Resource.__name__}"
        for resource, _ in cls._resources.values():
            if issubclass(resource, cls):
                # If a Resource inherits from the resource decorator, the server goes into an infinite recursion.
                # Here we make sure the user gets a clear error message (https://github.com/inmanta/inmanta-core/issues/8817).
                fq_name_current_resource = f"{resource.__module__}.{resource.__name__}"
                raise inmanta.ast.RuntimeException(
                    stmt=None,
                    msg=(
                        f"Resource {fq_name_current_resource} is inheriting from the {fq_name_resource_decorator} decorator."
                        f" Did you intend to inherit from {fq_name_resource_class} instead?"
                    ),
                )
            resource.validate()

    @classmethod
    def get_entity_resources(cls) -> Iterable[str]:
        """
        Returns a list of entity types for which a resource has been registered
        """
        return cls._resources.keys()

    @classmethod
    def get_class(cls, name: str) -> tuple[Optional[type["Resource"]], Optional[dict[str, str]]]:
        """
        Get the class definition for the given entity.
        """
        if name in cls._resources:
            return cls._resources[name]

        return (None, None)

    @classmethod
    def get_resources(cls) -> Iterator[tuple[str, type["Resource"]]]:
        """Return an iterator over resource type, resource definition"""
        return (
            (resource_type, resource_definition) for resource_type, (resource_definition, _options) in cls._resources.items()
        )

    @classmethod
    def reset(cls) -> None:
        """
        Clear the list of registered resources
        """
        cls._resources = {}


class ResourceNotFoundExcpetion(Exception):
    """
    This exception is thrown when a resource is not found
    """


@stable_api
class IgnoreResourceException(Exception):
    """
    Throw this exception when a resource should not be included by the exported.
    Typically resources use this to indicate that they are not managed by the orchestrator.
    """


def to_id(entity: "proxy.DynamicProxy") -> Optional[str]:
    """
    Convert an entity instance from the model to its resource id
    """
    entity_name = str(entity._type())
    for cls_name in [entity_name] + entity._type().get_all_parent_names():
        cls, options = resource.get_class(cls_name)

        if cls is not None:
            break

    if cls is not None and options is not None:
        obj_id = cls.object_to_id(entity, cls_name, options["name"], options["agent"])
        return str(obj_id)

    return None


class ResourceMeta(type):
    @classmethod
    def _get_parent_fields(cls, bases: Sequence[type["Resource"]]) -> list[str]:
        fields: list[str] = []
        for base in bases:
            if "fields" in base.__dict__:
                if not isinstance(base.__dict__["fields"], (tuple, list)):
                    raise Exception("fields attribute of %s should be a tuple or list" % base)

                fields.extend(base.__dict__["fields"])

            fields.extend(cls._get_parent_fields(base.__bases__))

        return fields

    def __new__(cls, class_name, bases, dct):
        fields = cls._get_parent_fields(bases)
        if "fields" in dct:
            if not isinstance(dct["fields"], (tuple, list)):
                raise Exception("fields attribute of %s should be a tuple or list" % class_name)

            fields.extend(dct["fields"])

        dct["fields"] = tuple(set(fields))
        return type.__new__(cls, class_name, bases, dct)


RESERVED_FOR_RESOURCE = {"id", "version", "model", "requires", "unknowns", "set_version", "clone", "is_type", "serialize"}


class ReferenceSubCollector:
    """
    A collector for references that:
    1. keeps track of all references it has seen
    2. keeps track of the paths at which these references have been seen
    """

    def __init__(self) -> None:
        self.references: dict[uuid.UUID, references.ReferenceModel] = {}
        self.replacements: dict[str, references.ReferenceModel] = {}

    def collect_reference(self, value: object) -> None:
        """Add a value reference and recursively add any other references."""
        match value:
            case list():
                for v in value:
                    self.collect_reference(v)

            case dict():
                for k, v in value.items():
                    self.collect_reference(v)

            case references.Reference():
                ref = value.serialize()
                self.references[ref.id] = ref
                for arg in value.arguments.values():
                    self.collect_reference(arg)

            case _:
                pass

    def add_reference(self, path: str, reference: "references.Reference[references.PrimitiveTypes]") -> None:
        """Add a new attribute map to a value reference that we found at the given path.

        :param path: The path where the value needs to be inserted
        :param reference: The attribute reference
        """
        self.collect_reference(reference)
        self.replacements[path] = reference.serialize()

    def collect_references(self, value: object, path: str) -> object:
        """
        Collect value references. This method also ensures that there are no values in the resources that are not serializable.
        This includes:
            - Unknowns
            - DynamicProxy


        :param value: The value to recursively find value references on
        :param path: The current path we are working on in the tree
        """

        def allow_references[T](v: T) -> T:
            if isinstance(v, proxy.DynamicProxy):
                # fails on stable mypy, but runs fine on master
                return v._allow_references()
            return v

        match value:
            case list() | proxy.SequenceProxy():
                return [
                    self.collect_references(value, f"{path}[{index}]") for index, value in enumerate(allow_references(value))
                ]

            case dict() | proxy.DictProxy():
                return {
                    key: self.collect_references(value, f"{path}.'{key.replace("'", "\\'")}'")
                    for key, value in allow_references(value).items()
                }

            case references.Reference():
                self.add_reference(path, value)
                return None

            case proxy.DynamicProxy() | util.Unknown():
                raise TypeError(f"{value!r} in resource is not JSON serializable at path {path}")

            case _:
                return value


class ReferenceCollector(ReferenceSubCollector):
    """Collect and organize all references and mutators for a specific resource"""

    def __init__(self, resource: "Resource") -> None:
        super().__init__()
        self.mutators: list[references.MutatorModel] = []
        self.resource = resource

    def add_reference(self, path: str, reference: "references.Reference[references.PrimitiveTypes]") -> None:
        """Add a new attribute map to a value reference that we found at the given path.

        :param path: The path where the value needs to be inserted
        :param reference: The attribute reference
        """
        super().add_reference(path, reference)
        self.mutators.append(
            references.ReplaceValue(
                resource=self.resource,
                value=reference,
                destination=path,
            ).serialize()
        )


@stable_api
class Resource(metaclass=ResourceMeta):
    """
    Plugins should inherit resource from this class so a resource from a model can be serialized and deserialized.

    Such as class is registered when the :func:`~inmanta.resources.resource` decorator is used. Each class needs to indicate
    the fields the resource will have with a class field named "fields". A metaclass merges all fields lists from the class
    itself and all superclasses. If a field it not available directly in the model object the serializer will look for
    static methods in the class with the name "get_$fieldname".
    """

    fields: Sequence[str] = (
        const.RESOURCE_ATTRIBUTE_SEND_EVENTS,
        const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS,
        const.RESOURCE_ATTRIBUTE_REFERENCES,
        const.RESOURCE_ATTRIBUTE_MUTATORS,
    )
    send_event: bool  # Deprecated field
    model: "proxy.DynamicProxy"
    map: dict[str, Callable[[Optional["export.Exporter"], "proxy.DynamicProxy"], Any]]

    @staticmethod
    def get_send_event(_exporter: "export.Exporter", obj: "Resource") -> bool:
        try:
            return obj.send_event
        except Exception:
            return False

    @staticmethod
    def get_references(_exporter: "export.Exporter", instance: "Resource") -> "typing.Sequence[references.ReferenceModel]":
        """This method is present so the serialization works correctly. This field is set by the serializer"""
        return []

    @staticmethod
    def get_mutators(_exporter: "export.Exporter", instance: "Resource") -> "typing.Sequence[references.MutatorModel]":
        """This method is present so the serialization works correctly. This field is set by the serializer"""
        return []

    @staticmethod
    def get_receive_events(_exporter: "export.Exporter", obj: "Resource") -> bool:
        try:
            return obj.receive_events
        except Exception:
            # default to True for backward compatibility (all resources used to receive events)
            return True

    @classmethod
    def convert_requires(
        cls, resources: dict["runtime.Instance", "Resource"], ignored_resources: set["runtime.Instance"]
    ) -> None:
        """
        Convert all requires

        :param resources: A dict with a mapping from model objects to resource objects
        :param ignored_resources: A set of model objects that have been ignored (and not converted to resources)
        """
        for res in resources.values():
            final_requires: set["Resource"] = set()
            initial_requires: list["runtime.Instance"] = [x for x in res.model.requires]

            for r in initial_requires:
                if r in resources:
                    final_requires.add(resources[r])

            if len(final_requires) == 0 and not len(initial_requires) == 0:
                for r in initial_requires:
                    # do not warn about resources that either contain unknowns or are ignored
                    if r in ignored_resources:
                        initial_requires.remove(r)

                if len(initial_requires) > 0:
                    LOGGER.warning(
                        "The resource %s had requirements before flattening, but not after flattening."
                        " Initial set was %s. Perhaps provides relation is not wired through correctly?",
                        res,
                        initial_requires,
                    )

            res.resource_requires = final_requires
            res.requires = {x.id for x in final_requires}

    @classmethod
    def object_to_id(
        cls, model_object: "proxy.DynamicProxy", entity_name: str, attribute_name: str, agent_attribute: str
    ) -> "Id":
        """
        Convert the given object to a textual id

        :param model_object: The object to convert to an id
        :param entity_name: The entity type
        :param attribute_name: The name of the attribute that uniquely identifies the entity
        :param agent_attribute: The "path" to the attribute that defines the agent
        """
        # first get the agent attribute
        path_elements: list[str] = agent_attribute.split(".")
        agent_value = model_object
        for i, el in enumerate(path_elements):
            try:
                # TODO cleanup this hack
                if isinstance(agent_value, list):
                    agent_value = agent_value[0]

                agent_value = getattr(agent_value, el)

            except inmanta.ast.UnexpectedReference:
                # Clean up exception
                current_path: str = ".".join(path_elements[: i + 1])
                raise ResourceException(
                    "Encountered reference in resource's agent attribute. Agent attribute values can not be references."
                    f" Encountered at attribute {current_path!r} of resource instance {model_object}"
                )
            except (ResourceException, inmanta.ast.UnsetException, inmanta.ast.UnknownException):
                raise
            except Exception:
                raise Exception(
                    "Unable to get the name of agent %s belongs to. In path %s, '%s' does not exist"
                    % (model_object, agent_attribute, el)
                )
        try:
            attribute_value = cls.map_field(None, entity_name, attribute_name, model_object)
        except inmanta.ast.UnexpectedReference:
            # Clean up exception
            raise ResourceException(
                "Encountered reference in resource's id attribute. Id attribute values can not be references."
                f" Encountered at attribute {attribute_name!r} of resource instance {model_object}"
            )
        if isinstance(attribute_value, util.Unknown):
            raise inmanta.ast.UnknownException(attribute_value)
        if not isinstance(agent_value, str):
            raise ResourceException(
                f"The agent attribute should lead to a string, got {agent_value} of type {type(agent_value)}"
            )

        # agent_value is no longer a proxy.DynamicProxy here, force this for mypy validation
        return Id(entity_name, str(agent_value), attribute_name, str(attribute_value))

    @classmethod
    def map_field(
        cls,
        exporter: Optional["export.Exporter"],
        entity_name: str,
        field_name: str,
        model_object: "proxy.DynamicProxy",
        reference_collector: Optional[ReferenceCollector] = None,
    ) -> object:
        """
        Map a field name to its value.

        :param reference_collector: Collector for references. If None, references may be returned, even if they can not
            be serialized later on. The caller is responsible for validating appropriately.
        """
        try:
            if hasattr(cls, "get_" + field_name):
                mthd = getattr(cls, "get_" + field_name)
                value = mthd(exporter, model_object)
            elif hasattr(cls, "map") and field_name in cls.map:
                value = cls.map[field_name](exporter, model_object)
            else:
                value = getattr(model_object, field_name)

            # walk the entire model to find any value references. Additionally we also want to make sure we raise exceptions on:
            # - Unknowns
            # - DynamicProxys to entities
            # - References if we don't have a reference_collector
            if reference_collector is not None:
                value = reference_collector.collect_references(value, field_name)

            return value
        except IgnoreResourceException:
            raise  # will be handled in _load_resources of export.py
        except inmanta.ast.UnknownException as e:
            return e.unknown
        except inmanta.ast.PluginException as e:
            raise inmanta.ast.ExplicitPluginException(
                None, f"Failed to get attribute '{field_name}' for export on '{entity_name}'", e
            )
        except inmanta.ast.CompilerException:
            # Internal exceptions (like UnsetException) should be propagated without being wrapped
            # as they are used later on and wrapping them would break the compiler
            raise
        except Exception as e:
            raise inmanta.ast.ExternalException(
                None, f"Failed to get attribute '{field_name}' for export on '{entity_name}'", e
            )

    @classmethod
    def create_from_model(cls, exporter: "export.Exporter", entity_name: str, model_object: "proxy.DynamicProxy") -> "Resource":
        """
        Build a resource from a given configuration model entity
        """
        resource_cls, options = resource.get_class(entity_name)

        if resource_cls is None or options is None:
            raise TypeError("No resource class registered for entity %s" % entity_name)

        # build the id of the object
        obj_id = resource_cls.object_to_id(model_object, entity_name, options["name"], options["agent"])
        obj = resource_cls(obj_id)

        with proxy.exportcontext:
            # map all fields
            reference_collector = ReferenceCollector(obj)
            fields: dict[str, object] = {
                field: resource_cls.map_field(exporter, entity_name, field, model_object, reference_collector)
                for field in resource_cls.fields
            }

        fields[const.RESOURCE_ATTRIBUTE_REFERENCES] = list(reference_collector.references.values())
        fields[const.RESOURCE_ATTRIBUTE_MUTATORS] = reference_collector.mutators

        obj.populate(fields)
        obj.model = model_object

        return obj

    @classmethod
    def deserialize(cls, obj_map: JsonType) -> "Resource":
        """Deserialize the resource from the given dictionary

        :param obj_map: The json structure that represents all fields of the resource
        """
        obj_id = Id.parse_id(obj_map["id"])
        cls_resource, _options = resource.get_class(obj_id.entity_type)

        force_fields = False
        if cls_resource is None:
            raise TypeError("No resource class registered for entity %s" % obj_id.entity_type)

        # backward compatibility for resources that were exported and stored in serialized form before these fields were
        # introduced:
        # - receive_events
        # - references
        # - mutators
        extra: dict[str, object] = {}
        if const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS not in obj_map:
            extra[const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS] = True

        if (
            const.RESOURCE_ATTRIBUTE_MUTATORS not in obj_map
            or obj_map[const.RESOURCE_ATTRIBUTE_MUTATORS] is None
            or const.RESOURCE_ATTRIBUTE_REFERENCES not in obj_map
            or obj_map[const.RESOURCE_ATTRIBUTE_REFERENCES] is None
        ):
            extra[const.RESOURCE_ATTRIBUTE_MUTATORS] = []
            extra[const.RESOURCE_ATTRIBUTE_REFERENCES] = []

        if extra:
            obj_map = {**obj_map, **extra}

        obj = cls_resource(obj_id)
        obj.populate(obj_map, force_fields)
        return obj

    @classmethod
    def validate(cls) -> None:
        for field in cls.fields:
            if field.startswith("_"):
                raise ResourceException("Resource field names can not start with _, reported in %s" % cls.__name__)
            if field in RESERVED_FOR_RESOURCE:
                raise ResourceException(
                    f"Resource {field} is a reserved keyword and not a valid field name, reported in {cls.__name__}"
                )

    def __init__(self, _id: "Id") -> None:
        self.id = _id
        self.requires: set[Id] = set()
        self.resource_requires: set[Resource] = set()
        self.unknowns: set[str] = set()

        if not hasattr(self.__class__, "fields"):
            raise Exception("A resource should have a list of fields")

        for field in self.__class__.fields:
            setattr(self, field, None)

        self.version = _id.version

        self._references_model: dict[uuid.UUID, references.ReferenceModel] = {}
        self._references: dict[uuid.UUID, references.Reference[references.RefValue]] = {}
        self._resolved = False

    def get(self, key: str, default: object = None) -> object:
        if key in self.fields:
            return getattr(self, key)
        return default

    def __getitem__(self, key: str) -> object:
        """Support dict like access on the resource"""
        if key in self.fields:
            return getattr(self, key)

        raise KeyError()

    def __contains__(self, item: str) -> object:
        return item in self.fields

    def __setitem__(self, key: str, value: object) -> None:
        """Support dict like access on the resource. It is not possible to create new
        attributes using setitem
        """
        if key in self.fields:
            return setattr(self, key, value)

        raise KeyError()

    def __delitem__(self, key: str) -> None:
        raise Exception("Deleting fields is not allowed on a resource")

    def clear(self) -> None:
        raise Exception("Deleting fields is not allowed on a resource")

    def items(self) -> Iterator[tuple[str, object]]:
        for key in self.fields:
            yield key, getattr(self, key)

    def get_reference_value(self, id: uuid.UUID, logger: "handler.LoggerABC") -> "references.RefValue":
        """Get a value of a reference"""
        if id not in self._references:
            if id not in self._references_model:
                raise KeyError(f"The reference with id {id} is not defined in resource {self.id}")

            model = self._references_model[id]
            ref = references.reference.get_class(model.type).deserialize(model, self, logger)
            self._references[model.id] = ref

        return self._references[id].get(logger)

    def resolve_all_references(self, logger: "handler.LoggerABC") -> None:
        """Resolve all value references"""
        if self._resolved:
            return
        for ref in self.references:  # type: ignore
            if ref.get("id", None) not in self._references_model:
                model = references.ReferenceModel(**ref)
                self._references_model[model.id] = model

        for mutator in self.mutators:  # type: ignore
            mutator = references.mutator.get_class(mutator["type"]).deserialize(
                references.MutatorModel(**mutator), self, logger
            )
            mutator.run(logger)

        self._resolved = True

    def populate(self, fields: dict[str, object], force_fields: bool = False) -> None:
        for field in self.__class__.fields:
            if field in fields or force_fields:
                setattr(self, field, fields[field])
            else:
                raise Exception("Resource with id {} does not have field {}".format(fields["id"], field))

        if "requires" in fields:
            # parse requires into ID's
            for require in fields["requires"]:  # type: ignore
                self.requires.add(Id.parse_id(require))

    def set_version(self, version: int) -> None:
        """Set the version of this resource"""
        self.version = version
        self.id.version = version

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, util.Unknown):
            self.unknowns.add(name)

        self.__dict__[name] = value

    def __str__(self) -> str:
        return str(self.id)

    def __repr__(self) -> str:
        return str(self)

    def clone[R: Resource](self: R, **kwargs: Any) -> R:
        """
        Create a clone of this resource. The given kwargs can be used to override attributes.

        :return: The cloned resource
        """
        res = Resource.deserialize(self.serialize())
        for k, v in kwargs.items():
            setattr(res, k, v)

        # Resources are copied in the hanlder, the cache has to be retained!
        res._resolved = self._resolved
        # In every case, share caches with child
        res._references = self._references
        res._references_model = self._references_model

        return res

    def serialize(self) -> JsonType:
        """
        Serialize this resource to its dictionary representation
        """
        dictionary: dict[str, Any] = {}

        for field in self.__class__.fields:
            dictionary[field] = getattr(self, field)

        dictionary["requires"] = [str(x) for x in self.requires]
        dictionary["version"] = self.version
        dictionary["id"] = self.id.resource_version_str()

        return dictionary

    def is_type(self, type_name: str) -> bool:
        return str(self.model._get_instance().get_type()) == type_name


@stable_api
class PurgeableResource(Resource):
    """
    See :inmanta:entity:`std::PurgeableResource` for more information.
    """

    fields = ("purged", "purge_on_delete")
    purged: bool
    purge_on_delete: bool


@stable_api
class DiscoveryResource(Resource):
    """
    See :inmanta:entity:`std::DiscoveryResource` for more information.
    """

    fields = ()


@stable_api
class ManagedResource(Resource):
    """
    See :inmanta:entity:`std::ManagedResource` for more information.
    """

    fields = ("managed",)

    managed: bool

    @staticmethod
    def get_managed(exporter: "export.Exporter", obj: "ManagedResource") -> bool:
        if not obj.managed:
            raise IgnoreResourceException()
        return obj.managed


PARSE_ID_REGEX = re.compile(
    r"^(?P<id>(?P<type>(?P<ns>[\w-]+(::[\w-]+)*)::(?P<class>[\w-]+))\[(?P<hostname>[^,]+),"
    r"(?P<attr>[^=]+)=(?P<value>[^\]]+)\])(,v=(?P<version>[0-9]+))?$"
)

PARSE_RVID_REGEX = re.compile(
    r"^(?P<id>(?P<type>(?P<ns>[\w-]+(::[\w-]+)*)::(?P<class>[\w-]+))\[(?P<hostname>[^,]+),"
    r"(?P<attr>[^=]+)=(?P<value>[^\]]+)\]),v=(?P<version>[0-9]+)$"
)


@stable_api
class Id:
    """
    A unique id that identifies a resource that is managed by an agent
    """

    def __init__(self, entity_type: str, agent_name: str, attribute: str, attribute_value: str, version: int = 0) -> None:
        """
        :attr entity_type: The resource type, as defined in the configuration model.
            For example :inmanta:entity:`std::testing::NullResource`.
        :attr agent_name: The agent responsible for this resource.
        :attr attribute: The key attribute that uniquely identifies this resource on the agent
        :attr attribute_value: The corresponding value for this key attribute.
        :attr version: The version number for this resource.
        """
        self._entity_type = entity_type
        self._agent_name = agent_name
        self._attribute = attribute
        self._attribute_value = attribute_value
        self._version = version

    def to_dict(self) -> JsonType:
        return {
            "entity_type": self._entity_type,
            "agent_name": self.agent_name,
            "attribute": self.attribute,
            "attribute_value": self.attribute_value,
            "version": self.version,
        }

    def get_entity_type(self) -> str:
        return self._entity_type

    def get_agent_name(self) -> str:
        return self._agent_name

    def get_attribute(self) -> str:
        return self._attribute

    def get_attribute_value(self) -> str:
        return self._attribute_value

    def get_version(self) -> int:
        return self._version

    def set_version(self, version: int) -> None:
        if self._version > 0 and version != self._version:
            raise AttributeError("can't set attribute version")

        self._version = version

    def get_inmanta_module(self) -> str:
        """
        Utility method to parse the Inmanta module out of
        the entity type for this Id.

        e.g. Returns `std` for resources of type `std::testing::NullResource`
        """
        ns = self._entity_type.split("::", maxsplit=1)
        return ns[0]

    def copy(self, *, version: int) -> "Id":
        """
        Creates a copy of this resource id for another model version.
        """
        return Id(self.entity_type, self.agent_name, self.attribute, self.attribute_value, version)

    def __str__(self) -> str:
        if self._version > 0:
            return self.resource_version_str()
        return self.resource_str()

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        return str(self) == str(other) and type(self) is type(other)

    def resource_str(self) -> ResourceIdStr:
        """
        String representation for this resource id with the following format:
            <type>[<agent>,<attribute>=<value>]

            - type: The resource type, as defined in the configuration model.
                For example :inmanta:entity:`std::testing::NullResource`.
            - agent: The agent responsible for this resource.
            - attribute: The key attribute that uniquely identifies this resource on the agent
            - value: The corresponding value for this key attribute.

        :return: Returns a :py:class:`inmanta.data.model.ResourceIdStr`
        """
        return cast(
            ResourceIdStr,
            "%(type)s[%(agent)s,%(attribute)s=%(value)s]"
            % {
                "type": self._entity_type,
                "agent": self._agent_name,
                "attribute": self._attribute,
                "value": self._attribute_value,
            },
        )

    def resource_version_str(self) -> ResourceVersionIdStr:
        return cast(
            ResourceVersionIdStr,
            "%(type)s[%(agent)s,%(attribute)s=%(value)s],v=%(version)s"
            % {
                "type": self._entity_type,
                "agent": self._agent_name,
                "attribute": self._attribute,
                "value": self._attribute_value,
                "version": self._version,
            },
        )

    def __repr__(self) -> str:
        return str(self)

    def get_instance(self) -> Optional[Resource]:
        """
        Create an instance of this class and set the identifying attribute already
        """
        cls, _ = resource.get_class(self.entity_type)

        if cls is None:
            return None

        obj = cls(self)

        setattr(obj, self.attribute, self.attribute_value)
        return obj

    @classmethod
    def parse_resource_version_id(cls, resource_id: ResourceVersionIdStr) -> "Id":
        id: Id = Id.parse_id(resource_id)
        if id.version == 0:
            raise ValueError(f"Version is missing from resource id: {resource_id}")
        return id

    @classmethod
    def parse_id(cls, resource_id: Union[ResourceVersionIdStr, ResourceIdStr], version: Optional[int] = None) -> "Id":
        """
        Parse the resource id and return the type, the hostname and the
        resource identifier.

        :param version: If provided, the version field of the returned Id will be set to this version.
        """
        result = PARSE_ID_REGEX.search(resource_id)

        if result is None:
            raise ValueError("Invalid id for resource %s" % resource_id)

        if version is None:
            version_match: str = result.group("version")

            version = 0
            if version_match is not None:
                version = int(version_match)

        id_obj = Id(result.group("type"), result.group("hostname"), result.group("attr"), result.group("value"), version)
        return id_obj

    @classmethod
    def is_resource_version_id(cls, value: str) -> bool:
        """
        Check whether the given value is a resource version id
        """
        result = PARSE_RVID_REGEX.search(value)
        return result is not None

    @classmethod
    def is_resource_id(cls, value: str) -> bool:
        """
        Check whether the given value is a resource id
        """
        result = PARSE_ID_REGEX.search(value)
        return result is not None

    def is_resource_version_id_obj(self) -> bool:
        """
        Check whether this object represents a resource version id
        """
        return self.version != 0

    @classmethod
    def set_version_in_id(cls, id_str: Union[ResourceVersionIdStr, ResourceIdStr], new_version: int) -> ResourceVersionIdStr:
        """
        Return a copy of the given id_str with the version number set to new_version.
        """
        parsed_id = cls.parse_id(id_str)
        new_id = parsed_id.copy(version=new_version)
        return new_id.resource_version_str()

    entity_type = property(get_entity_type)
    agent_name = property(get_agent_name)
    attribute = property(get_attribute)
    attribute_value = property(get_attribute_value)
    version = property(get_version, set_version)


class HostNotFoundException(Exception):
    """
    This exception is raise when the deployment agent cannot access a host to manage a resource (Use mainly with remote io)
    """

    def __init__(self, hostname: str, user: str, error: str) -> None:
        self.hostname = hostname
        self.user = user
        self.error = error

    def to_action(self) -> "ResourceAction":
        from inmanta.data import ResourceAction

        ra = ResourceAction()  # @UndefinedVariable
        ra.message = f"Failed to access host {self.hostname} as user {self.user} over ssh."
        ra.data = {"host": self.hostname, "user": self.user, "error": self.error}

        return ra



"""
Copyright 2016 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import base64
import logging
import os
import uuid
from collections import defaultdict
from pathlib import Path
from typing import cast

import pytest

import inmanta.server.services.environmentlistener
from inmanta.data import model
from inmanta.module import ModuleLoadingException, Project
from inmanta.server import SLICE_ENVIRONMENT
from inmanta.server.services import environmentservice
from utils import log_contains


async def test_project_api_v1(client):
    result = await client.create_project("project-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]

    project_id = result.result["project"]["id"]

    result = await client.create_project("project-test")
    assert result.code == 500

    result = await client.list_projects()
    assert result.code == 200
    assert "projects" in result.result
    assert len(result.result["projects"]) == 1

    assert result.result["projects"][0]["id"] == project_id

    result = await client.get_project(id=project_id)
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]["id"] == project_id
    assert result.result["project"]["name"] == "project-test"

    result = await client.modify_project(id=project_id, name="project-test2")
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]["id"] == project_id
    assert result.result["project"]["name"] == "project-test2"

    result = await client.get_project(id=project_id)
    assert result.code == 200
    assert "project" in result.result
    assert result.result["project"]["id"] == project_id
    assert result.result["project"]["name"] == "project-test2"

    result = await client.delete_project(id=project_id)
    assert result.code == 200

    result = await client.list_projects()
    assert result.code == 200
    assert "projects" in result.result
    assert len(result.result["projects"]) == 0

    # get non existing environment
    response = await client.get_environment(uuid.uuid4())
    assert response.code == 404


async def test_delete_environment_project(client):
    """Test modifying the project of an environment"""

    # Create two projects and two environments
    result = await client.create_project("project-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]
    project_id_a = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id_a, name="env")
    assert result.code == 200
    env1_id = result.result["environment"]["id"]

    # Delete project a
    result = await client.delete_project(project_id_a)
    assert result.code == 409
    assert (
        result.result["message"] == "Request conflicts with the current state of the resource: "
        f"Cannot remove the project `{project_id_a}` because it still contains some environments: "
        f"('env', '{env1_id}')"
    )

    result = await client.delete_environment(id=env1_id)
    assert result.code == 200

    result = await client.delete_project(project_id_a)
    assert result.code == 200

    result = await client.list_projects()
    assert result.code == 200
    assert result.result["projects"] == []


async def test_project_api_v2_project_list_ordering(client_v2):
    """
    Creates a few projects with several environments each.
    Check that they are ordered by ascending (project_name, environment_name)
    """

    project_environments_map: dict[str, list[str]] = defaultdict(list)

    for project_n in range(3):
        project_name: str = f"test-project-{project_n}"
        result = await client_v2.project_create(project_name)
        assert result.code == 200
        assert "data" in result.result
        assert "id" in result.result["data"]

        project_id: uuid.UUID = result.result["data"]["id"]
        project_environments_map[project_name] = []

        for environment_n in range(3):
            env_name: str = f"test-env-{environment_n}"
            result = await client_v2.environment_create(project_id=project_id, name=env_name)
            assert result.code == 200
            project_environments_map[project_name].append(env_name)

    result = await client_v2.project_list()
    assert result.code == 200
    assert "data" in result.result
    assert len(result.result["data"]) == 3
    for i in range(3):
        assert len(result.result["data"][i]["environments"]) == 3

    # Make sure results are sorted according to project name...
    assert sorted(project_environments_map.keys()) == [result.result["data"][i]["name"] for i in range(3)]

    # ... and according to env name within each project
    for project_n, key_value_pair in enumerate(sorted(project_environments_map.items())):
        project_id, env_id_list = key_value_pair
        assert sorted(env_id_list) == [env["name"] for env in result.result["data"][project_n]["environments"]]


async def test_project_api_v2(client_v2):
    result = await client_v2.project_create("project-test")
    assert result.code == 200
    assert "data" in result.result
    assert "id" in result.result["data"]

    project_id = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id, name="dev")
    assert result.code == 200
    assert "data" in result.result
    assert "id" in result.result["data"]
    assert "project_id" in result.result["data"]
    assert project_id == result.result["data"]["project_id"]
    assert "dev" == result.result["data"]["name"]
    env1_id = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id, name="dev2")
    assert result.code == 200
    assert "data" in result.result
    assert "id" in result.result["data"]
    assert "project_id" in result.result["data"]
    assert project_id == result.result["data"]["project_id"]
    assert "dev2" == result.result["data"]["name"]

    # modify branch and repo
    result = await client_v2.environment_modify(id=env1_id, name="dev", repository="test")
    assert result.code == 200

    result = await client_v2.environment_modify(id=env1_id, name="dev", branch="test")
    assert result.code == 200

    result = await client_v2.project_list()
    assert result.code == 200
    assert "data" in result.result
    assert len(result.result["data"]) == 1
    assert len(result.result["data"][0]["environments"]) == 2

    # Failure conditions

    # Delete non existing project
    response = await client_v2.project_delete(uuid.uuid4())
    assert response.code == 404

    # Modify non existing project
    response = await client_v2.project_modify(uuid.uuid4(), name="test")
    assert response.code == 404

    # Modify to duplicate name
    response = await client_v2.project_create("project2")
    assert response.code == 200

    response = await client_v2.project_modify(project_id, name="project2")
    assert response.code == 500

    # Get non existing project
    response = await client_v2.project_get(uuid.uuid4())
    assert response.code == 404

    # Create env in non existing project
    result = await client_v2.environment_create(project_id=uuid.uuid4(), name="dev")
    assert result.code == 404

    # Create a duplicate environment
    result = await client_v2.environment_create(project_id=project_id, name="dev")
    assert result.code == 400

    # Modify a non existing environment
    result = await client_v2.environment_modify(id=uuid.uuid4(), name="dev")
    assert result.code == 404

    # Create a duplicate environment
    result = await client_v2.environment_create(project_id=project_id, name="dev", repository="")
    assert result.code == 400

    # Modify to duplicate environment
    result = await client_v2.environment_modify(id=env1_id, name="dev2")
    assert result.code == 400

    # Get an environment
    result = await client_v2.environment_get(id=env1_id)
    assert result.code == 200
    assert result.result["data"]["name"] == "dev"

    # Operation on non existing
    result = await client_v2.environment_get(id=uuid.uuid4())
    assert result.code == 404

    result = await client_v2.environment_delete(id=uuid.uuid4())
    assert result.code == 404


async def test_modify_environment_project(client_v2):
    """Test modifying the project of an environment"""

    # Create two projects and two environments
    result = await client_v2.project_create("dev-project")
    assert result.code == 200
    project_id_a = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id_a, name="env")
    assert result.code == 200
    env1_id = result.result["data"]["id"]

    result = await client_v2.project_create("test-project")
    assert result.code == 200

    project_id_b = result.result["data"]["id"]
    result = await client_v2.environment_create(project_id=project_id_b, name="env")
    assert result.code == 200

    # Try to move environment from project a to b, but the name in that project is not free
    result = await client_v2.environment_modify(id=env1_id, name="env", project_id=project_id_b)
    assert result.code == 400

    # Move and rename environment from project a to b
    result = await client_v2.environment_modify(id=env1_id, name="new-env", project_id=project_id_b)
    assert result.code == 200
    assert result.result["data"]["project_id"] == project_id_b

    # Delete project a
    response = await client_v2.project_delete(project_id_a)
    assert response.code == 200

    # The environment should still exist
    result = await client_v2.environment_get(id=env1_id)
    assert result.code == 200

    # Try to move it back to the not existing project
    result = await client_v2.environment_modify(id=env1_id, name="new-env", project_id=project_id_a)
    assert result.code == 400

    # Make sure that specifying the current project id does not cause problems
    result = await client_v2.environment_modify(id=env1_id, name="new-env", project_id=project_id_b)
    assert result.code == 200


async def test_delete_environment_project_v2(client_v2):
    """Test deleting a project containing two environments."""

    # Create a project and two environments
    result = await client_v2.project_create("dev-project")
    assert result.code == 200
    project_id_a = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id_a, name="env")
    assert result.code == 200
    env1_id = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id_a, name="env2")
    assert result.code == 200
    env2_id = result.result["data"]["id"]

    # Delete project
    result = await client_v2.project_delete(project_id_a)
    assert result.code == 409
    assert (
        result.result["message"] == "Request conflicts with the current state of the resource: "
        f"Cannot remove the project `{project_id_a}` because it still contains some environments: "
        f"('env', '{env1_id}'),('env2', '{env2_id}')"
    )

    result = await client_v2.environment_delete(id=env1_id)
    assert result.code == 200

    result = await client_v2.environment_delete(id=env2_id)
    assert result.code == 200

    result = await client_v2.project_delete(project_id_a)
    assert result.code == 200

    result = await client_v2.project_list()
    assert result.code == 200
    assert result.result["data"] == []


async def test_env_api(client):
    result = await client.create_project("env-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    assert result.code == 200
    assert "environment" in result.result
    assert "id" in result.result["environment"]
    assert "project" in result.result["environment"]
    assert project_id == result.result["environment"]["project"]
    assert "dev" == result.result["environment"]["name"]

    env_id = result.result["environment"]["id"]

    result = await client.modify_environment(id=env_id, name="dev2")
    assert result.code == 200
    assert "environment" in result.result
    assert result.result["environment"]["id"] == env_id
    assert result.result["environment"]["name"] == "dev2"

    result = await client.get_environment(id=env_id)
    assert result.code == 200
    assert "environment" in result.result
    assert result.result["environment"]["id"] == env_id
    assert result.result["environment"]["project"] == project_id
    assert result.result["environment"]["name"] == "dev2"

    project_result = await client.get_project(id=project_id)
    assert project_result.code == 200
    assert "project" in project_result.result
    assert env_id in project_result.result["project"]["environments"]

    result = await client.list_environments()
    assert result.code == 200
    assert len(result.result) == 1

    result = await client.delete_environment(id=env_id)
    assert result.code == 200

    result = await client.list_environments()
    assert result.code == 200
    assert "environments" in result.result
    assert len(result.result["environments"]) == 0


async def test_create_env_same_name(client):
    result = await client.create_project("env-test")
    assert result.code == 200
    assert "project" in result.result
    assert "id" in result.result["project"]
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev1")
    assert result.code == 200

    result = await client.create_environment(project_id=project_id, name="dev2")
    assert result.code == 200
    env_id = result.result["environment"]["id"]

    # try to create a new env with the same name as an existing one
    result = await client.create_environment(project_id=project_id, name="dev1")
    assert result.code == 400
    assert f"Project with id={project_id} already has an environment with name dev1" in result.result["message"]

    # try to rename an existing env using a name that is already used.
    result = await client.modify_environment(id=env_id, name="dev1")
    assert result.code == 400
    assert f"Project with id={project_id} already has an environment with name dev1" in result.result["message"]


async def test_project_list_environments_after_failed_removal(client):
    result = await client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    result = await client.create_environment(project_id=project_id, name="prod")

    result = await client.delete_project(project_id)
    assert result.code == 409

    result = await client.list_environments()
    assert len(result.result["environments"]) == 2


async def test_create_with_id(client):
    project_id = uuid.uuid4()
    result = await client.create_project(name="test_project", project_id=project_id)
    assert result.result["project"]["id"] == str(project_id)

    env_id = uuid.uuid4()
    result = await client.create_environment(project_id=project_id, name="test_env", environment_id=env_id)
    env = result.result["environment"]
    assert env["id"] == str(env_id)
    assert env["project"] == str(project_id)

    result = await client.create_environment(project_id=project_id, name="test_env2", environment_id=env_id)
    assert result.code == 500


@pytest.mark.parametrize("no_agent", [True])
async def test_environment_listener(server, client_v2, caplog):
    class EnvironmentListenerCounter(inmanta.server.services.environmentlistener.EnvironmentListener):
        def __init__(self):
            self.created_counter = 0
            self.updated_counter = 0
            self.cleared_counter = 0
            self.deleted_counter = 0

        async def environment_action_cleared(self, env: model.Environment) -> None:
            self.cleared_counter += 1

        async def environment_action_created(self, env: model.Environment) -> None:
            self.created_counter += 1
            if self.created_counter == 3:
                raise Exception("Something is not right")

        async def environment_action_deleted(self, env: model.Environment) -> None:
            self.deleted_counter += 1

        async def environment_action_updated(self, updated_env: model.Environment, original_env: model.Environment) -> None:
            self.updated_counter += 1

    environment_listener = EnvironmentListenerCounter()

    environment_service = cast(environmentservice.EnvironmentService, server.get_slice(SLICE_ENVIRONMENT))
    environment_service.register_listener_for_multiple_actions(
        environment_listener,
        {
            inmanta.server.services.environmentlistener.EnvironmentAction.created,
            inmanta.server.services.environmentlistener.EnvironmentAction.updated,
            inmanta.server.services.environmentlistener.EnvironmentAction.deleted,
            inmanta.server.services.environmentlistener.EnvironmentAction.cleared,
        },
    )
    result = await client_v2.project_create("project-test")
    assert result.code == 200

    project_id = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id, name="dev")
    assert result.code == 200
    env1_id = result.result["data"]["id"]

    result = await client_v2.environment_create(project_id=project_id, name="dev2")
    assert result.code == 200

    # modify branch and repo
    result = await client_v2.environment_modify(id=env1_id, name="dev", repository="test")
    assert result.code == 200

    result = await client_v2.environment_modify(id=env1_id, name="dev", branch="test")
    assert result.code == 200

    result = await client_v2.project_list()
    assert result.code == 200

    # Get an environment
    result = await client_v2.environment_get(id=env1_id)
    assert result.code == 200
    assert result.result["data"]["name"] == "dev"

    result = await client_v2.environment_delete(id=uuid.uuid4())
    assert result.code == 404

    # Clear
    result = await client_v2.environment_clear(id=env1_id)
    assert result.code == 200

    # Delete
    result = await client_v2.environment_delete(id=env1_id)
    assert result.code == 200

    result = await client_v2.environment_create(project_id=project_id, name="dev3")
    assert result.code == 200

    log_contains(
        caplog,
        "inmanta.server.services.environmentservice",
        logging.WARNING,
        "Notifying listener of created failed with the following exception",
    )

    assert environment_listener.created_counter == 3
    assert environment_listener.updated_counter == 2
    assert environment_listener.cleared_counter == 1
    assert environment_listener.deleted_counter == 1


@pytest.mark.parametrize("install", [True, False])
def test_project_load_install(snippetcompiler_clean, local_module_package_index: str, install: bool) -> None:
    """
    Verify that loading a project only installs modules when install is True.
    """
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import minimalv2module",
        python_requires=[inmanta.util.parse_requirement("inmanta-module-minimalv2module")],
        install_project=False,
        index_url=local_module_package_index,
    )
    if install:
        project.load(install=True)
    else:
        with pytest.raises(ModuleLoadingException, match="Failed to load module minimalv2module"):
            project.load()
        # make sure project load works after installing modules
        project.install_modules()
        project.load()


@pytest.fixture
def environment_icons() -> dict[str, str]:
    icon_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "icon")
    icon_names = ["logo.jpeg", "logo.png", "logo.svg", "logo.webp"]
    icon_dict = {}
    for name in icon_names:
        icon_dict[name.split(".")[1]] = base64.b64encode(Path(os.path.join(icon_dir, name)).read_bytes()).decode("utf-8")
    return icon_dict


async def test_environment_icon_description(client_v2, environment_icons: dict[str, str]):
    """Test creating an environment with an icon and description"""

    result = await client_v2.project_create("dev-project")
    assert result.code == 200
    project_id_a = result.result["data"]["id"]
    desc = "This is an environment"
    result = await client_v2.environment_create(project_id=project_id_a, name="env", description=desc)
    assert result.code == 200
    assert result.result["data"]["description"] == desc

    # Test description length
    result = await client_v2.environment_create(project_id=project_id_a, name="env2", description="a" * 256)
    assert result.code == 400
    result = await client_v2.environment_create(project_id=project_id_a, name="env2", description="a" * 255)
    assert result.code == 200

    for image_type, image in environment_icons.items():
        mime_type = image_type if image_type != "svg" else "svg+xml"
        result = await client_v2.environment_create(
            project_id=project_id_a, name=f"env_{image_type}", description=desc, icon=f"image/{mime_type};base64,{image}"
        )
        assert result.code == 200

    raw_icon = environment_icons["svg"]

    # Test invalid icon data strings
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xm;base64,{raw_icon}"
    )
    assert result.code == 400

    # Specify an icon with an invalid base64 encoding
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml;base64,{raw_icon[0:10]}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml;,{raw_icon}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml;{raw_icon}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml;base64;{raw_icon}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml,base64;{raw_icon}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml,base64,{raw_icon}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xmlbase64{raw_icon}"
    )
    assert result.code == 400
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/svg+xml;base64{raw_icon}"
    )
    assert result.code == 400

    # Check too large icon, base64 encoding increases the size
    large_icon = base64.b64encode(b"a" * 65535).decode("utf-8")
    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon=f"image/png;base64,{large_icon}"
    )
    assert result.code == 400

    result = await client_v2.environment_create(
        project_id=project_id_a, name="envx", description=desc, icon="image/png;base64,"
    )
    assert result.code == 400

    # Test modification of the description and icon
    name_of_env_to_modify = "env_no_icon"
    result = await client_v2.environment_create(project_id=project_id_a, name=name_of_env_to_modify, description=desc, icon="")
    assert result.code == 200
    id_of_env_to_modify = result.result["data"]["id"]

    icon_data_url = f"image/png;base64,{environment_icons['png']}"
    # Add an icon
    result = await client_v2.environment_modify(id_of_env_to_modify, name_of_env_to_modify, icon=icon_data_url)
    assert result.code == 200
    assert result.result["data"]["icon"] == icon_data_url

    new_description = "new desc"
    # Change the description, but keep the icon the same
    result = await client_v2.environment_modify(id_of_env_to_modify, name_of_env_to_modify, description=new_description)
    assert result.code == 200
    assert result.result["data"]["icon"] == icon_data_url

    # Make sure GET request works
    result = await client_v2.environment_get(id_of_env_to_modify, details=True)
    assert result.code == 200
    assert result.result["data"]["icon"] == icon_data_url
    assert result.result["data"]["description"] == new_description

    # Delete the icon
    result = await client_v2.environment_modify(id_of_env_to_modify, name_of_env_to_modify, icon="")
    assert result.code == 200

    result = await client_v2.environment_get(id_of_env_to_modify, details=True)
    assert result.code == 200
    assert result.result["data"]["icon"] == ""

    result = await client_v2.environment_modify(id_of_env_to_modify, name_of_env_to_modify, description="b" * 256)
    assert result.code == 400


async def test_environment_icon_with_details_only(client_v2, environment_icons: dict[str, str]):
    """Test that the icon for an environment is only returned when explicitly requested"""
    result = await client_v2.project_create("dev-project")
    assert result.code == 200
    project_id_a = result.result["data"]["id"]
    # Create environment with description and icon
    icon_data_string = f"image/png;base64,{environment_icons['png']}"
    description = "desc"
    result = await client_v2.environment_create(
        project_id=project_id_a, name="env", description=description, icon=icon_data_string
    )
    assert result.code == 200
    env_id = result.result["data"]["id"]
    # Check that the icon and description are not returned without the details flag
    result = await client_v2.environment_get(env_id)
    assert result.code == 200
    assert result.result["data"]["icon"] == ""
    assert result.result["data"]["description"] == ""
    # With the details, they should be returned correctly
    result = await client_v2.environment_get(env_id, details=True)
    assert result.code == 200
    assert result.result["data"]["icon"] == icon_data_string
    assert result.result["data"]["description"] == description

    result = await client_v2.environment_list()
    assert result.code == 200
    assert result.result["data"][0]["icon"] == ""
    assert result.result["data"][0]["description"] == ""

    result = await client_v2.environment_list(details=True)
    assert result.code == 200
    assert result.result["data"][0]["icon"] == icon_data_string
    assert result.result["data"][0]["description"] == description

    result = await client_v2.project_get(project_id_a)
    assert result.code == 200
    assert result.result["data"]["environments"][0]["icon"] == ""
    assert result.result["data"]["environments"][0]["description"] == ""

    result = await client_v2.project_get(project_id_a, environment_details=True)
    assert result.code == 200
    assert result.result["data"]["environments"][0]["icon"] == icon_data_string
    assert result.result["data"]["environments"][0]["description"] == description

    result = await client_v2.project_list()
    assert result.code == 200
    assert result.result["data"][0]["environments"][0]["icon"] == ""
    assert result.result["data"][0]["environments"][0]["description"] == ""

    result = await client_v2.project_list(environment_details=True)
    assert result.code == 200
    assert result.result["data"][0]["environments"][0]["icon"] == icon_data_string
    assert result.result["data"][0]["environments"][0]["description"] == description



from setuptools import setup, find_namespace_packages
from os import path

requires = [
    "asyncpg~=0.25",
    "build~=1.0",
    "click-plugins~=1.0",
    # click has been known to publish non-backwards compatible minors in the past (removed deprecated code in 8.1.0)
    "click>=8.0,<8.4",
    "colorlog~=6.4",
    "cookiecutter>=1,<3",
    "crontab>=0.23,<2.0",
    "cryptography>=36,<47",
    # docstring-parser has been known to publish non-backwards compatible minors in the past
    "docstring-parser>=0.10,<0.18",
    "email-validator>=1,<3",
    "jinja2~=3.0",
    "more-itertools>=8,<11",
    # upper bound on packaging because we use a non-public API that might change in any (non-SemVer) version
    "packaging>=21.3,<25.1",
    # pip>=21.3 required for editable pyproject.toml + setup.cfg based install support
    "pip>=21.3",
    "ply~=3.0",
    "pydantic~=2.5,!=2.9.2",
    "PyJWT~=2.0",
    "pynacl~=1.5",
    "python-dateutil~=2.0",
    "pyyaml~=6.0",
    "setuptools",
    "texttable~=1.0",
    # tornado>6.5 because of CVE https://github.com/advisories/GHSA-7cx3-6m66-7c5m
    "tornado>6.5",
    # lower bound because of ilevkivskyi/typing_inspect#100
    "typing_inspect~=0.9",
    "ruamel.yaml~=0.17",
    "toml~=0.10 ",
    "setproctitle~=1.3",
    "jsonpath-ng~=1.7",
]


# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# This version is managed by bumpversion. Should you ever update it manually, make sure to consistently update it everywhere
# (See the bumpversion.cfg file for relevant locations).
version = "15.4.0"

setup(
    version=version,
    python_requires=">=3.12",  # also update classifiers
    # Meta data
    name="inmanta-core",
    description="Inmanta deployment tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Inmanta",
    author_email="code@inmanta.com",
    url="https://github.com/inmanta/inmanta-core",
    license="Apache Software License 2",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="orchestrator orchestration configurationmanagement",
    project_urls={
        "Bug Tracker": "https://github.com/inmanta/inmanta-core/issues",
        "Documentation": "https://docs.inmanta.com/community/latest/",
    },
    # Packaging
    package_dir={"": "src"},
    # All data files should be treated as namespace package according to
    # https://setuptools.pypa.io/en/latest/userguide/datafiles.html#subdirectory-for-data-files
    packages=find_namespace_packages(where="src"),
    # https://www.python.org/dev/peps/pep-0561/#packaging-type-information
    zip_safe=False,
    include_package_data=True,
    install_requires=requires,
    extras_require={
        "dev": [
            # all extra's (for testing and mypy)
            "inmanta-core[datatrace,debug,tracing]",
            # test dependencies
            "inmanta-dev-dependencies[pytest,async,core]",
            "inmanta-module-std",
            "bumpversion",
            "openapi_spec_validator",
            "pep8-naming",
            "pip2pi",
            "psutil",
            "time-machine",
            # types
            "types-python-dateutil",
            "types-PyYAML",
            "types-setuptools",
            "types-toml",
            # doc dependencies
            "furo",
            "inmanta-sphinx",
            "myst-parser",
            "sphinx",
            "sphinx-argparse",
            "sphinx-autodoc-annotation",
            "sphinx-click",
            "sphinxcontrib-contentui",
            "sphinxcontrib.datatemplates",
            "sphinxcontrib-redoc",
            "sphinxcontrib-serializinghtml",
            "sphinx-design",
            "Sphinx-Substitution-Extensions",
        ],
        "debug": ["rpdb"],
        # option to install a matched pair of inmanta-core and pytest-inmanta-extensions
        "pytest-inmanta-extensions": [f"pytest-inmanta-extensions~={version}.0.dev"],
        "datatrace": ["graphviz"],
        "tracing": ["logfire>=0.46,<5.0", "opentelemetry-instrumentation-asyncpg~=0.46b0"],
    },
    entry_points={
        "console_scripts": [
            "inmanta-cli = inmanta.main:main",
            "inmanta = inmanta.app:app",
            "inmanta-initial-user-setup = inmanta.user_setup:main",
        ],
        "inmanta.mypy.methods": [
            "methods_v2 = inmanta.protocol.methods_v2",
            "methods = inmanta.protocol.methods",
        ]
    },
)



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import base64
import binascii
import logging
import os
import re
import shutil
import uuid
import warnings
from collections import defaultdict
from collections.abc import Set
from concurrent import futures
from re import Pattern
from typing import Optional, cast

import asyncpg
from asyncpg import StringDataRightTruncationError

from inmanta import config, data
from inmanta.data import AUTOSTART_AGENT_DEPLOY_INTERVAL, AUTOSTART_AGENT_REPAIR_INTERVAL, Setting, model
from inmanta.protocol import encode_token, handle, methods, methods_v2
from inmanta.protocol.common import ReturnValue, attach_warnings
from inmanta.protocol.exceptions import BadRequest, Forbidden, NotFound, ServerError
from inmanta.server import (
    SLICE_AGENT_MANAGER,
    SLICE_AUTOSTARTED_AGENT_MANAGER,
    SLICE_COMPILER,
    SLICE_DATABASE,
    SLICE_ENVIRONMENT,
    SLICE_ORCHESTRATION,
    SLICE_RESOURCE,
    SLICE_SERVER,
    SLICE_TRANSPORT,
    agentmanager,
    protocol,
)
from inmanta.server.server import Server
from inmanta.server.services import orchestrationservice, resourceservice
from inmanta.server.services.environmentlistener import (  # These were moved from this module, important to keep them in place
    EnvironmentAction,
    EnvironmentListener,
)
from inmanta.types import Apireturn, JsonType, Warnings

LOGGER = logging.getLogger(__name__)


def rename_fields(env: model.Environment) -> JsonType:
    env_dict = env.model_dump()
    env_dict["project"] = env_dict["project_id"]
    del env_dict["project_id"]
    return env_dict


class EnvironmentService(protocol.ServerSlice):
    """Slice with project and environment management"""

    server_slice: Server
    agent_manager: "agentmanager.AgentManager"
    autostarted_agent_manager: "agentmanager.AutostartedAgentManager"
    orchestration_service: "orchestrationservice.OrchestrationService"
    resource_service: "resourceservice.ResourceService"
    listeners: dict[EnvironmentAction, list[EnvironmentListener]]
    # environment_state_operation_lock is to prevent concurrent execution of
    # operations that modify the state of an environment, such as halting, resuming, or deleting.
    # This lock helps prevent race conditions and ensures that state changes are carried out in a safe and
    # sequential manner. It guarantees that operations affecting the agent states and environment status
    # do not overlap.
    environment_state_operation_lock: asyncio.Lock
    icon_regex: Pattern[str] = re.compile("^(image/png|image/jpeg|image/webp|image/svg\\+xml);(base64),(.+)$")

    def __init__(self) -> None:
        super().__init__(SLICE_ENVIRONMENT)
        self.listeners = defaultdict(list)
        self.environment_state_operation_lock = asyncio.Lock()
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=1)

    def get_dependencies(self) -> list[str]:
        return [
            SLICE_COMPILER,
            SLICE_SERVER,
            SLICE_DATABASE,
            SLICE_AUTOSTARTED_AGENT_MANAGER,
            SLICE_ORCHESTRATION,
            SLICE_RESOURCE,
            SLICE_AGENT_MANAGER,
        ]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    async def prestart(self, server: protocol.Server) -> None:
        await super().prestart(server)
        self.server_slice = cast(Server, server.get_slice(SLICE_SERVER))
        self.agent_manager = cast(agentmanager.AgentManager, server.get_slice(SLICE_AGENT_MANAGER))
        self.autostarted_agent_manager = cast(
            agentmanager.AutostartedAgentManager, server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
        )
        self.orchestration_service = cast(orchestrationservice.OrchestrationService, server.get_slice(SLICE_ORCHESTRATION))
        self.resource_service = cast(resourceservice.ResourceService, server.get_slice(SLICE_RESOURCE))
        # Register the compiler service here to the environment service listener. Registering it within the compiler service
        # would result in a circular dependency between the environment slice and the compiler service slice.
        self.register_listener_for_multiple_actions(
            self.server_slice.compiler, {EnvironmentAction.cleared, EnvironmentAction.deleted}
        )

    async def start(self) -> None:
        await super().start()
        await self._enable_schedules_all_envs()

    async def _enable_schedules_all_envs(self) -> None:
        """
        Schedules appropriate actions for schedule-related settings for all environments. Overrides old schedules.
        """
        env: data.Environment
        for env in await data.Environment.get_list(details=False):
            await self._enable_schedules(env)

    async def _enable_schedules(self, env: data.Environment, setting: Optional[data.Setting] = None) -> None:
        """
        Schedules appropriate actions for a single environment according to the settting value. Overrides old schedules.

        :param setting: Only schedule appropriate actions for this setting, if any.
        """
        if setting is None or setting.name == data.AUTO_FULL_COMPILE:
            if setting is not None:
                LOGGER.info("Environment setting %s changed. Rescheduling full compiles.", setting.name)
            self.server_slice.compiler.schedule_full_compile(env, str(await env.get(data.AUTO_FULL_COMPILE)))

    def _disable_schedules(self, env: data.Environment) -> None:
        """
        Removes scheduling of all appropriate actions for a single environment.
        """
        self.server_slice.compiler.schedule_full_compile(env, schedule_cron="")

    async def _setting_change(self, env: data.Environment, key: str) -> Warnings:
        setting = env._settings[key]

        warnings = None
        if setting.recompile:
            LOGGER.info("Environment setting %s changed. Recompiling with update = %s", key, setting.update)
            metadata = model.ModelMetadata(
                message="Recompile for modified setting", type="setting", extra_data={"setting": key}
            )
            warnings = await self.server_slice._async_recompile(env, setting.update, metadata=metadata.model_dump())

        if setting.agent_restart:
            LOGGER.info("Environment setting %s changed. Restarting agents.", key)
            self.add_background_task(self.autostarted_agent_manager.restart_agents(env))

        if key in [
            AUTOSTART_AGENT_DEPLOY_INTERVAL,
            AUTOSTART_AGENT_REPAIR_INTERVAL,
        ]:
            await self.autostarted_agent_manager.notify_agent_deploy_timer_update(env)

        self.add_background_task(self._enable_schedules(env, setting))

        return warnings

    # v1 handlers
    @handle(methods.create_environment)
    async def create_environment(
        self, project_id: uuid.UUID, name: str, repository: str, branch: str, environment_id: Optional[uuid.UUID]
    ) -> Apireturn:
        return (
            200,
            {"environment": rename_fields(await self.environment_create(project_id, name, repository, branch, environment_id))},
        )

    @handle(methods.modify_environment, environment_id="id")
    async def modify_environment(self, environment_id: uuid.UUID, name: str, repository: str, branch: str) -> Apireturn:
        return 200, {"environment": rename_fields(await self.environment_modify(environment_id, name, repository, branch))}

    @handle(methods.get_environment, environment_id="id", api_version=1)
    async def get_environment(
        self, environment_id: uuid.UUID, versions: Optional[int] = None, resources: Optional[int] = None
    ) -> Apireturn:
        versions = 0 if versions is None else int(versions)
        resources = 0 if resources is None else int(resources)

        env = await data.Environment.get_by_id(environment_id)

        if env is None:
            raise NotFound("The environment id does not exist.")

        env_dict = env.to_dict()

        if versions > 0:
            env_dict["versions"] = await data.ConfigurationModel.get_versions(environment_id, limit=versions)

        if resources > 0:
            env_dict["resources"] = await data.Resource.get_resources_report(environment=environment_id)

        return 200, {"environment": env_dict}

    @handle(methods.list_environments)
    async def list_environments(self) -> Apireturn:
        return 200, {"environments": [rename_fields(env) for env in await self.environment_list()]}

    @handle(methods.delete_environment, environment_id="id")
    async def delete_environment(self, environment_id: uuid.UUID) -> Apireturn:
        await self.environment_delete(environment_id)
        return 200

    @handle(methods_v2.halt_environment, env="tid")
    async def halt(self, env: data.Environment, connection: Optional[asyncpg.connection.Connection] = None) -> None:
        async with self.environment_state_operation_lock:
            await self._halt(env, connection)

    async def _halt(
        self,
        env: data.Environment,
        connection: Optional[asyncpg.connection.Connection] = None,
        *,
        delete_agent_venv: bool = False,
    ) -> None:
        """
        Halts the specified environment without acquiring the environment_state_operation_lock.
        This method is designed to be an internal helper that allows for halting an environment
        as part of a larger operation (e.g., deletion), where the lock is managed by the caller and prevent double locking.

        :param delete_agent_venv: True iff also delete the venv of the agent after stopping it.
        """
        async with data.Environment.get_connection(connection=connection) as con:
            async with con.transaction():
                refreshed_env: Optional[data.Environment] = await data.Environment.get_by_id(env.id, connection=con)
                if refreshed_env is None:
                    raise NotFound("Environment %s does not exist" % env.id)

                # silently ignore requests if this environment has already been halted
                if refreshed_env.halted:
                    return

                await refreshed_env.update_fields(halted=True, connection=con)
                await self.agent_manager.halt_agents(refreshed_env, connection=con)
        await self.autostarted_agent_manager.stop_agents(refreshed_env, delete_venv=delete_agent_venv)

    @handle(methods_v2.resume_environment, env="tid")
    async def resume(self, env: data.Environment) -> None:
        async with self.environment_state_operation_lock:
            await self._resume(env)

    async def _resume(self, env: data.Environment, connection: Optional[asyncpg.connection.Connection] = None) -> None:
        """
        Internal helper to resume an environment. This method must be called under the self.environment_state_operation_lock.
        """
        async with data.Environment.get_connection(connection) as con:
            async with con.transaction():
                refreshed_env: Optional[data.Environment] = await data.Environment.get_by_id(env.id, connection=con)
                if refreshed_env is None:
                    raise NotFound("Environment %s does not exist" % env.id)
                if refreshed_env.is_marked_for_deletion:
                    raise BadRequest("Cannot resume an environment that is marked for deletion.")
                # silently ignore requests if this environment has already been resumed
                if not refreshed_env.halted:
                    return

                await refreshed_env.update_fields(halted=False, connection=con)
                await self.agent_manager.resume_agents(refreshed_env, connection=con)
        await self.autostarted_agent_manager.restart_agents(refreshed_env)
        await self.server_slice.compiler.resume_environment(refreshed_env.id)

    @handle(methods.clear_environment, env="id")
    async def clear_environment(self, env: data.Environment) -> Apireturn:
        await self.environment_clear(env)
        return 200

    @handle(methods.create_token, env="tid")
    async def create_token(self, env: data.Environment, client_types: list[str], idempotent: bool) -> Apireturn:
        """
        Create a new auth token for this environment
        """
        return 200, {"token": await self.environment_create_token(env, client_types, idempotent)}

    @handle(methods.list_settings, env="tid")
    async def list_settings(self, env: data.Environment) -> Apireturn:
        setting_details: dict[str, model.EnvironmentSettingDetails] = {
            k: env.settings.get(k) for k in sorted(env.settings.settings.keys()) if k in data.Environment._settings.keys()
        }
        settings = {setting_name: details.value for setting_name, details in setting_details.items()}
        setting_definitions = dict(sorted(data.Environment.get_setting_definitions_for_api(setting_details).items()))
        return 200, {"settings": settings, "metadata": setting_definitions}

    @handle(methods.set_setting, env="tid", key="id")
    async def set_setting(self, env: data.Environment, key: str, value: model.EnvSettingType) -> Apireturn:
        if env.settings.is_protected(key):
            raise Forbidden(
                f"Cannot update environment setting {key} because it's protected"
                f" (reason={env.settings.get_protected_by_description(key)})."
            )
        try:
            original_env = env.to_dto()
            await env.set(key, value)
            warnings = await self._setting_change(env, key)
            await self.notify_listeners(EnvironmentAction.updated, env.to_dto(), original_env)
            return attach_warnings(200, None, warnings)
        except KeyError:
            raise NotFound()
        except ValueError as e:
            raise BadRequest(f"Invalid value. {e}")

    @handle(methods.get_setting, env="tid", key="id")
    async def get_setting(self, env: data.Environment, key: str) -> Apireturn:
        setting = await self.environment_setting_get(env, key)
        return 200, {"value": setting.settings[key], "metadata": setting.definition}

    @handle(methods.delete_setting, env="tid", key="id")
    async def delete_setting(self, env: data.Environment, key: str) -> Apireturn:
        if env.settings.is_protected(key):
            raise Forbidden(
                f"Cannot delete environment setting {key} because it's protected"
                f" (reason={env.settings.get_protected_by_description(key)})."
            )
        try:
            original_env = env.to_dto()
            await env.unset(key)
            warnings = await self._setting_change(env, key)
            await self.notify_listeners(EnvironmentAction.updated, env.to_dto(), original_env)
            return attach_warnings(200, None, warnings)
        except KeyError:
            raise NotFound()

    # v2 handlers
    @handle(methods_v2.environment_create)
    async def environment_create(
        self,
        project_id: uuid.UUID,
        name: str,
        repository: str,
        branch: str,
        environment_id: Optional[uuid.UUID],
        description: str = "",
        icon: str = "",
    ) -> model.Environment:
        # check if an environment with this name is already defined in this project
        envs = await data.Environment.get_list(project=project_id, name=name)
        if len(envs) > 0:
            raise BadRequest(f"Project with id={project_id} already has an environment with name {name}")

        if environment_id is None:
            environment_id = uuid.uuid4()

        if (repository is None and branch is not None) or (repository is not None and branch is None):
            raise BadRequest("Repository and branch should be set together.")
        if repository is None:
            repository = ""
        if branch is None:
            branch = ""

        # fetch the project first
        project = await data.Project.get_by_id(project_id)
        if project is None:
            raise NotFound("The project id for the environment does not exist.")

        # check if an environment with this name is already defined in this project
        envs = await data.Environment.get_list(project=project_id, name=name)
        if len(envs) > 0:
            raise ServerError(f"Project {project.name} (id={project.id}) already has an environment with name {name}")

        self.validate_icon(icon)

        env = data.Environment(
            id=environment_id,
            name=name,
            project=project_id,
            repo_url=repository,
            repo_branch=branch,
            description=description,
            icon=icon,
        )
        try:
            await env.insert()
        except StringDataRightTruncationError:
            raise BadRequest("Maximum size of the icon data url or the description exceeded")
        await self.notify_listeners(EnvironmentAction.created, env.to_dto())
        await self._enable_schedules(env)
        return env.to_dto()

    def validate_icon(self, icon: str) -> None:
        """Check if the icon is in the supported format, raise an exception otherwise"""
        if icon == "":
            return
        match = self.icon_regex.match(icon)
        if match and len(match.groups()) == 3:
            encoded_image = match.groups()[2]
            try:
                base64.b64decode(encoded_image, validate=True)
            except binascii.Error:
                raise BadRequest("The icon is not a valid base64 encoded string")
        else:
            raise BadRequest("The value supplied for the icon parameter is invalid")

    @handle(methods_v2.environment_modify, environment_id="id")
    async def environment_modify(
        self,
        environment_id: uuid.UUID,
        name: str,
        repository: Optional[str],
        branch: Optional[str],
        project_id: Optional[uuid.UUID] = None,
        description: Optional[str] = None,
        icon: Optional[str] = None,
    ) -> model.Environment:
        env = await data.Environment.get_by_id(environment_id)
        if env is None:
            raise NotFound("The environment id does not exist.")
        original_env = env.to_dto()

        project = project_id or env.project

        # check if an environment with this name is already defined in this project
        envs = await data.Environment.get_list(project=project, name=name)
        if len(envs) > 0 and envs[0].id != environment_id:
            raise BadRequest(f"Project with id={project} already has an environment with name {name}")

        fields: dict[str, str | int | uuid.UUID] = {"name": name}
        if repository is not None:
            fields["repo_url"] = repository

        if branch is not None:
            fields["repo_branch"] = branch

        # Update the project field if requested and the project exists
        if project_id is not None:
            project_from_db = await data.Project.get_by_id(project_id)
            if not project_from_db:
                raise BadRequest(f"Project with id={project_id} doesn't exist")
            fields["project"] = project_id

        if description is not None:
            fields["description"] = description

        if icon is not None:
            self.validate_icon(icon)
            fields["icon"] = icon

        try:
            await env.update_fields(connection=None, **fields)
        except StringDataRightTruncationError:
            raise BadRequest("Maximum size of the icon data url or the description exceeded")
        await self.notify_listeners(EnvironmentAction.updated, env.to_dto(), original_env)
        return env.to_dto()

    @handle(methods_v2.environment_get, environment_id="id", api_version=2)
    async def environment_get(self, environment_id: uuid.UUID, details: bool = False) -> model.Environment:
        env = await data.Environment.get_by_id(environment_id, details=details)

        if env is None:
            raise NotFound("The environment id does not exist.")

        return env.to_dto()

    @handle(methods_v2.environment_list)
    async def environment_list(self, details: bool = False) -> list[model.Environment]:
        # data access framework does not support multi-column order by, but multi-environment projects are rare
        # (and discouraged)
        # => sort by primary column in SQL, then do full sort in Python, cheap because mostly sorted already by this point
        env_list = await data.Environment.get_list(details=details, order_by_column="project")
        return sorted((env.to_dto() for env in env_list), key=lambda e: (e.project_id, e.name, e.id))

    @handle(methods_v2.environment_delete, environment_id="id")
    async def environment_delete(self, environment_id: uuid.UUID) -> None:
        async with self.environment_state_operation_lock:
            async with data.Environment.get_connection() as connection:
                env = await data.Environment.get_by_id(environment_id, connection=connection)
                if env is None:
                    raise NotFound("The environment with given id does not exist.")

                is_protected_environment = await env.get(data.PROTECTED_ENVIRONMENT, connection)
                if is_protected_environment:
                    raise Forbidden(
                        f"Environment {environment_id} is protected. See environment setting: {data.PROTECTED_ENVIRONMENT}"
                    )

                await env.mark_for_deletion(connection=connection)

                # Check if the environment is halted; if not, halt it
                if not env.halted:
                    LOGGER.info("Halting Environment %s", str(environment_id))
                    await self._halt(env, connection=connection, delete_agent_venv=True)

                self._disable_schedules(env)
                await self.server_slice.compiler.cancel_compile(env.id)
                # Delete the environment directory before deleting the database records. This ensures that
                # this operation can be retried if the deletion of the environment directory fails. Otherwise,
                # the environment directory would be left in an inconsistent state. This can cause problems if
                # the user later on recreates an environment with the same environment id.
                await self._delete_environment_dir(environment_id)
                await env.delete_cascade(connection=connection)

            await self.notify_listeners(EnvironmentAction.deleted, env.to_dto())

    @handle(methods_v2.environment_clear, env="id")
    async def environment_clear(self, env: data.Environment) -> None:
        """
        Clear the environment
        """
        async with self.environment_state_operation_lock:
            async with data.Environment.get_connection() as connection:
                is_protected_environment = await env.get(data.PROTECTED_ENVIRONMENT, connection=connection)
                if is_protected_environment:
                    raise Forbidden(f"Environment {env.id} is protected. See environment setting: {data.PROTECTED_ENVIRONMENT}")

                initial_halted_state = env.halted

                # Check if the environment is halted; if not, halt it
                if not env.halted:
                    LOGGER.info("Halting environment %s for clear operation", str(env.id))
                    await self._halt(env, connection=connection, delete_agent_venv=True)

                # Keep this method call under the self.environment_state_operation_lock lock, because cancel_compile()
                # must be called on halted environments only.
                await self.server_slice.compiler.cancel_compile(env.id)
                # Delete the environment directory before deleting the database records. This ensures that
                # this operation can be retried if the deletion of the environment directory fails. Otherwise,
                # the environment directory would be left in an inconsistent state. This can cause problems if
                # the user later on recreates an environment with the same environment id.
                await self._delete_environment_dir(env.id)
                await env.clear(connection=connection)

                await self.notify_listeners(EnvironmentAction.cleared, env.to_dto())

                if not initial_halted_state:
                    # Make sure to restore the environment to its initial state
                    LOGGER.info("Resume environment %s because clear operation finished", str(env.id))
                    await self._resume(env, connection=connection)

    @handle(methods_v2.environment_create_token, env="tid")
    async def environment_create_token(self, env: data.Environment, client_types: list[str], idempotent: bool) -> str:
        """
        Create a new auth token for this environment
        """
        if not config.Config.getboolean("server", "auth", False):
            raise BadRequest("Authentication is disabled, generating a token is not allowed")
        return encode_token(client_types, str(env.id), idempotent)

    @handle(methods_v2.environment_settings_list, env="tid")
    async def environment_settings_list(self, env: data.Environment) -> model.EnvironmentSettingsReponse:
        setting_details: dict[str, model.EnvironmentSettingDetails] = {
            k: env.settings.get(k) for k in sorted(env.settings.settings.keys()) if k in data.Environment._settings.keys()
        }
        settings = {setting_name: details.value for setting_name, details in setting_details.items()}
        setting_definitions = dict(sorted(data.Environment.get_setting_definitions_for_api(setting_details).items()))
        return model.EnvironmentSettingsReponse(settings=settings, definition=setting_definitions)

    @handle(methods_v2.environment_settings_set, env="tid", key="id")
    async def environment_settings_set(self, env: data.Environment, key: str, value: model.EnvSettingType) -> ReturnValue[None]:
        if env.settings.is_protected(key):
            raise Forbidden(
                f"Cannot update environment setting {key} because it's protected"
                f" (reason={env.settings.get_protected_by_description(key)})."
            )
        try:
            original_env = env.to_dto()
            await env.set(key, value)
            warnings = await self._setting_change(env, key)
            result: ReturnValue[None] = ReturnValue(response=None)
            if warnings:
                result.add_warnings(warnings)
            await self.notify_listeners(EnvironmentAction.updated, env.to_dto(), original_env)
            return result
        except KeyError:
            raise NotFound()
        except ValueError:
            raise ServerError("Invalid value")

    @handle(methods_v2.environment_setting_get, env="tid", key="id")
    async def environment_setting_get(self, env: data.Environment, key: str) -> model.EnvironmentSettingsReponse:
        try:
            value = await env.get(key)
            setting_definitions = data.Environment.get_setting_definitions_for_api(env.settings.get_all())
            return model.EnvironmentSettingsReponse(
                settings={key: value},
                definition=setting_definitions,
            )
        except KeyError:
            raise NotFound()

    @handle(methods_v2.environment_setting_delete, env="tid", key="id")
    async def environment_setting_delete(self, env: data.Environment, key: str) -> ReturnValue[None]:
        if env.settings.is_protected(key):
            raise Forbidden(
                f"Cannot delete environment setting {key} because it's protected"
                f" (reason={env.settings.get_protected_by_description(key)})."
            )
        try:
            original_env = env.to_dto()
            await env.unset(key)
            warnings = await self._setting_change(env, key)
            result: ReturnValue[None] = ReturnValue(response=None)
            if warnings:
                result.add_warnings(warnings)
            await self.notify_listeners(EnvironmentAction.updated, env.to_dto(), original_env)
            return result
        except KeyError:
            raise NotFound()

    @handle(methods_v2.protected_environment_settings_set_batch, env="tid")
    async def protected_environment_settings_set_batch(
        self, env: data.Environment, settings: dict[str, model.EnvSettingType], protected_by: model.ProtectedBy
    ) -> None:
        await env.set_protected_environment_settings(protected_settings=settings, protected_by=protected_by)

    def register_listener_for_multiple_actions(
        self, current_listener: EnvironmentListener, actions: Set[EnvironmentAction]
    ) -> None:
        """
            Should only be called during pre-start
        :param current_listener: The listener to register
        :param actions: type of actions the listener is interested in
        """
        for action in actions:
            self.register_listener(current_listener, action)

    def register_listener(self, current_listener: EnvironmentListener, action: EnvironmentAction) -> None:
        """
            Should only be called during pre-start
        :param current_listener: The listener to register
        :param action: type of action the listener is interested in
        """
        self.listeners[action].append(current_listener)

    def remove_listener(self, action: EnvironmentAction, current_listener: EnvironmentListener) -> None:
        self.listeners[action].remove(current_listener)

    async def _delete_environment_dir(self, environment_id: uuid.UUID) -> None:
        """
        Deletes an environment from the server's state_dir directory. This method should be called after
        notify_listeners() to ensure that the listeners are notified.

        :param environment_id: The uuid of the environment to remove from the state directory.

        :raises ServerError: When a file or directory has been created by a user different than the
        one running the Inmanta server inside the environment directory marked for removal.
        """
        state_dir = config.state_dir.get()
        environment_dir = os.path.join(state_dir, "server", str(environment_id))

        if os.path.exists(environment_dir):
            loop = asyncio.get_running_loop()
            try:
                # This call might fail when someone manually creates a directory or file that is owned
                # by another user than the user running the inmanta server.
                await loop.run_in_executor(
                    self.thread_pool,
                    shutil.rmtree,
                    environment_dir,
                )
            except PermissionError:
                raise ServerError(
                    f"Environment {environment_id} cannot be deleted because it contains files owned"
                    " by a different user than the one running the Inmanta server."
                )

    async def notify_listeners(
        self,
        action: EnvironmentAction,
        updated_env: model.Environment,
        original_env: Optional[model.Environment] = None,
    ) -> None:
        for current_listener in self.listeners[action]:
            try:
                if action == EnvironmentAction.created:
                    await current_listener.environment_action_created(updated_env)
                if action == EnvironmentAction.deleted:
                    await current_listener.environment_action_deleted(updated_env)
                if action == EnvironmentAction.cleared:
                    await current_listener.environment_action_cleared(updated_env)
                if action == EnvironmentAction.updated and original_env:
                    await current_listener.environment_action_updated(updated_env, original_env)
            except Exception:
                LOGGER.warning("Notifying listener of %s failed with the following exception", action.value, exc_info=True)

    async def register_setting(self, setting: Setting) -> None:
        """
        Should only be called during pre-start
        Adds a new setting to the environments from outside inmanta-core.
        As example, inmanta-lsm can use this method to add settings that are only
        relevant for inmanta-lsm but that are needed in the environment.
        :param setting: the setting that should be added to the existing settings
        """
        warnings.warn(
            "Registering environment settings via the inmanta.server.services.environmentservice.register_setting endpoint "
            "is deprecated. Environment settings defined by an extension should be advertised via the "
            "register_environment_settings method of the inmanta_ext.<extension_name>.extension.py file of an extension.",
            category=DeprecationWarning,
        )
        data.Environment.register_setting(setting)
